import numpy as np
import lmfit
from copy import copy
import sys
import pandas as pd
import matplotlib
from matplotlib import pyplot as plt
import re

class MSD:

    def __init__(self, msd):

        self.t = np.ma.masked_invalid(msd[0])
        ind_sort = np.argsort(self.t)
        self.t = self.t[ind_sort]

        self.msd = np.ma.masked_invalid(msd[1,ind_sort])
        self.dmsd = np.ma.masked_invalid(msd[2,ind_sort])

        # properties
        self.fix = None

        # attributes defined in private methods
        self._lmpars = None
        self._weights = None
        self._varnames = None

        self._fit_data = None
        self._fit_weights = None
        self._fit_logarithmic = False

        # fit results
        self.pars = None
        self.fit_result = []

        self._minimizer = None

    @staticmethod
    def _reduce_func(arr):
        return -0.5 * np.nansum(arr * arr)

    @staticmethod
    def _iter_cb(params, iter, resid, *args, **kws):
        if (iter) % 4 == 0:
            print('\rOptimizing...(%d)' % iter, end='', flush=1)
        else:
            pass


    def fit(self, mode='sig', init={}, fix={}, lmfit_pars={}, fitqdep={}):
        '''
        Function that computes the fits using lmfit's minimizer
        '''

        self.fix = fix

        # make the parameters
        if not lmfit_pars.get('is_weighted', True):
            init['__lnsigma'] = 1

        self._init_parameters(init, fix)

        # setting the weights of the data
        self._fit_data = self.msd
        self._get_weights(mode)
        self._fit_weights = self._weights


        self._minimizer = lmfit.Minimizer(self._residuals, params=self._lmpars,
                                          reduce_fcn=self._reduce_func,
                                          iter_cb=self._iter_cb,
                                          nan_policy='omit')

        # do the fit
        out = self._minimizer.minimize(**lmfit_pars)

        self._write_to_pars(out,)
        self.fit_result.append((out, lmfit.fit_report(out)))

        # convert to numeric data types
        self.pars = self.pars.apply(pd.to_numeric)

        self._fit_logarithmic = False

        return self.pars, self.fit_result


    def plot(self, doplot=False, marker='o', ax=None, xl=None, yl=None, colors=None, alpha=1.,
             markersize=3., data_label=None, confint=False, pars=None, **kwargs):

        sumidit = True if (self.pars is not None) else False

        if pars is not None :
            if len(pars):
                self.pars = pars
                sumidit = True

        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(self.t)*0.5,np.max(self.t)*1.5)

        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]), 50)

        for ii, iq in enumerate([0]):
            pl = []
            if sumidit:
                msdf = self._calc_model(self.fit_result[ii][0].params, x=xf)

                labstr_fit = None
                if 'legf' in doplot and sumidit:
                    pard = {'t':r'$t: {:.2e}\mathrm{{s}},\,$',
                            'g':r'$\gamma: {:.2g},\,$',
                            'b':r'$\mathrm{{b}}: {:.3g},\,$',
                            'a':r'$\mathrm{{a}}: {:.3g},\,$'}
                    labstr_fit = ''
                    for i in range(self.nmodes):
                        for vn in 'tgba':
                            if vn == 'a' and i > 0:
                                continue
                            elif vn == 'a' and i == 0:
                                vnn = 'a'
                            else:
                                vnn = vn + str(i)
                            if vnn in self.pars.columns:
                                labstr_fit += pard[vn].format(self.pars.loc[ii, vnn])
                            elif fix is not None and vnn in fix.keys():
                                labstr_fit += 'fix '+ pard[vn].format(fix[vnn])
                            else:
                                labstr_fit += pard[vn].format(0)

            labstr_data = None
            if 'legd' in doplot:
                if data_label is not None:
                    labstr_data = data_label
                else:
                    labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'legq' in doplot:
                labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'fit' in doplot and sumidit:
                pl.append(ax.plot(xf, msdf, '-', label=labstr_fit, linewidth=1))

            if 'data' in doplot:
                pl.append(ax.errorbar(self.t.filled(np.nan), self.msd.filled(np.nan), yerr=self.dmsd.filled(np.nan),
                                          linestyle='', marker=marker, label=labstr_data,
                                          alpha=alpha, markersize=markersize)[0])

            if colors is None:
                if 'data' in doplot:
                    color = pl[-1].get_color()
                else:
                    color = 'gray'
            else:
                color = colors[ii]

            for p in pl:
                if isinstance(p, list):
                    p[0].set_color(color)
                elif isinstance(p, matplotlib.container.ErrorbarContainer):
                    for child in p.get_children():
                        child.set_color(color)

            ax.set_xscale('log')
            ax.set_yscale('log')

            if 'leg' in doplot:
                ax.legend()

            if 'report' in doplot and sumidit:
                print(lmfit.fit_report(out))

            ax.set_xlabel(r'$\tau\,(\mathrm{s})$')
            ax.set_ylabel(r'$g_2(\tau)$')

            ax.set_xlim(*xl)
            if yl is not None:
                ax.set_ylim(*yl)

        return None

    def _calc_model(self, v, x=None):
        """Calculate multi mode g2 funktion
        """
        t = self.t if x is None else x
        model = 6 * v['d']**2 * (1 - np.exp(-(v['D0']/v['d']**2 * t)**v['a'])) ** (1/v['a']) * (1 + v['Dm']/v['d']**2 * t)
        if self._fit_logarithmic:
            return np.log10(model)
        else:
            return model

    def _residuals(self, pars, *args, **kwargs):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        model = self._calc_model(v)

        resid = (self._fit_data - model) * np.abs(self._fit_weights)
        return resid


    def _init_parameters(self, init, fix):
        '''Initialize lmfit parameters dictionary
        '''

        self._make_varnames()
        self._initial_guess(init)

        # initialize parameters
        pars = lmfit.Parameters()
        for vn, vinit in init.items():
            pars.add(vn, value=vinit[0], min=vinit[1], max=vinit[2], vary=1)


        self._init_pars_dataframe(init)

        # setting parameters fixed
        for vn in fix.keys():
            if vn in pars.keys():
                pars[vn].set(value=fix[vn], min=-np.inf, max=np.inf, vary=0)

        # print(pars)
        self._lmpars = pars

    def _get_weights(self, mode):
        '''
        mask data points to exclude them from the fit based on error bars and nan values
        and define weights for the fit
        '''
        dmsd = self.dmsd
        msd = self.msd.copy()
        if dmsd is not None and mode is not 'none':
            excerr = (dmsd.filled(0)<=0)
            wgt = dmsd.copy()
            wgt = np.ma.masked_where(excerr, wgt)
            if mode == 'semilogx':
                wgt = np.log10(wgt)
            elif mode == 'semilogx2':
                wgt = 1/np.log10(wgt/msd)
            elif mode == 'semilogx3':
                wgt = 1/(msd * np.log10(wgt/msd))
            elif (mode == 'equal') or (mode == 'none') or (mode == None):
                wgt = np.ones_like(wgt)
            elif mode == 'logt':
                wgt = 1/np.log10(self.t)
            elif mode == 't':
                wgt = 1/self.t
            elif mode == 'sig**2':
                wgt = 1/wgt**2
            elif mode == 'sig':
                wgt = 1/wgt
            elif mode == 'logsig':
                wgt = 1/np.log10(wgt)
            elif mode == 'data':
                wgt = 1/self.msd.copy()
            elif mode == 'test':
                self._fit_data = np.log10(self._fit_data)
                wgt = 1/np.log10(wgt)
                self._fit_logarithmic = True
            else:
                raise ValueError(f'Error mode {mode} not understood.')
        else:
            wgt = np.ma.masked_array(np.ones_like(self.msd))

        excdat = ~np.isfinite(self.msd) | wgt.mask
        self.msd = np.ma.masked_where(excdat, self.msd)

        self._weights = wgt

    def _init_pars_dataframe(self, init):
        """
        initialize pandas DataFrame for returning results
        """
        names = list(init.keys())
        cols = [names[i // 2] if (i + 1) % 2 else 'd' + names[i // 2]
                for i in range(len(names) * 2)]
        cols.extend(['chisqr', 'redchi', 'bic', 'aic'])
        self.pars = pd.DataFrame(columns=cols)

    def _initial_guess(self, init):
        """
        make initial guess for parameters if not provided by init
        """
        for vn in self._varnames:
            if vn not in init.keys():
                if vn == 'd':
                    init[vn] = (50, 0, None)
                elif vn == 'Dm':
                    init[vn] = (1e8, 0, None)
                elif vn == 'D0':
                    init[vn] = (1e8, 0, None)
                elif vn == 'a':
                    init[vn] = (1, 0, None)

    def _make_varnames(self):
        """Make dict of varnames for easy handling in _calc_model
        """
        self._varnames = {'d': [],
                          'D0': [],
                          'Dm': [],
                          'a': []
                          }

    def _write_to_pars(self, out):
        """ Save fit results in self.pars variable
        """
        for k in self._varnames.keys():
            cond = k in ['__lnsigma']
            gof = np.hstack([out.chisqr, out.redchi, out.bic, out.aic])
            param_name = df_name = k
            value = out.params[param_name].value
            try:
                stderr = 1. * out.params[param_name].stderr
            except TypeError:
                stderr = np.nan
            self.pars.loc[0, df_name] = value
            self.pars.loc[0, 'd'+df_name] = stderr
            self.pars.iloc[0, -4:] = gof







#! /usr/bin/env python
import numpy as np
from matplotlib import pyplot as plt
from functools import wraps


class Decorators:
    """
    Decorators class: Decorators are defined here.
    """
    @staticmethod
    def input2list(func):
        @wraps(func)
        def wrapper(a, x, *args, **kwargs):
            if np.issubdtype(type(x), np.integer):
                if x == -1:
                    x = a.meta.index.values
                else:
                    x = [x,]
            return func(a, x, *args, **kwargs)
        return wrapper

    @staticmethod
    def init_figure(x=(1,1), y=(6,5)):
        def init_figure_decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if 'ax' in kwargs:
                    ax = kwargs['ax']
                    kwargs_new = dict(kwargs)
                    del kwargs_new['ax']
                else:
                    fig, ax = plt.subplots(*x, figsize=y)
                    if type(ax) == np.ndarray:
                        ax = ax.flatten()
                    kwargs_new = dict(kwargs)
                return func(*args, ax=ax, **kwargs_new)
            return wrapper
        return init_figure_decorator






import numpy as np
from ..Xplot.plotqrois import plotqrois

def check_dimension(setup, img):
    s = np.shape(img)
    if s != setup.detector.dim:
        setup.dim = s
        print(f'Detector dimensiion has been changed to {s}.')
    else:
        return None

def getqroi(saxs, setup, qr, phir=None, mirror=False):

    q = setup.ai.array_from_unit(unit='q_nm^-1')
    radius = setup.ai.array_from_unit(unit='r_m')/75e-6
    phi = setup.ai.chiArray()

    qv = qr[:,0]
    dqv = qr[:,1]

    if phir is None:
        phiv = [0.]
        dphi = [360.]
    else:
        phiv = phir[:,0]
        dphi = phir[:,1]

    phiv = phiv*np.pi/180
    dphi = dphi*np.pi/180

    ind = []
    r = []
    ph = []

    for i in range(len(qv)):
        for j in range(len(phiv)):
            tmp_q = (q>=(qv[i]-dqv[i]/2)) & (q<=(qv[i]+dqv[i]/2)) & setup.mask
            phit = phi.copy()
            phit = (phit - phiv[j]) % (2*np.pi)
            if mirror:
                tmp_phi = (phit <= dphi[j]) | (((phit - np.pi) % (2*np.pi)) <= dphi[j])
            else:
                tmp_phi = (phit<=(dphi[j]))
            tmp = np.where(tmp_q & tmp_phi)
            del phit
            if len(tmp[0]):
                ind.append(tmp)
                r_min = radius[tmp].min()
                r_max = radius[tmp].max()
                r.append((r_max, r_max - r_min))
    return ind, r

def flatten_init(inp):

    def convert(s):
        if np.issubdtype(type(s[0]), np.number):
            return (np.array([s[0]]), s[1])
        else:
            return s

    def get_stack(p):
        p = convert(p)
        return np.hstack((p[0],np.ones(len(p[0]))*p[1])).reshape(-1,2, order='F')

    i = 0
    for p in inp:
        stack = get_stack(p)
        if i == 0:
            rois = stack.copy()
        else:
            rois = np.vstack((rois,stack))
        i += 1
    return rois

def defineqrois(setup, Isaxs, qv_init=None, phiv_init=[(0,360)], 
                plot=False, d=250, mirror=False, **kwargs):

    # check_dimension(setup, Isaxs)

    if qv_init is None:
        try:
            qv_init = setup.qv_init
            if phiv_init is None and 'phiv_init' in vars(setup):
                phiv_init = setup.phiv_init
        except KeyError:
            print('Setup does not contain "qv_init" to initialize Q ROIs.')
    else:
        qv_init = flatten_init(qv_init)
        phiv_init = flatten_init(phiv_init)
        setup.qv_init = qv_init
        setup.phiv_init = phiv_init

    phiv_init[:,0] -= phiv_init[:,1]/2

    qroi, radii = getqroi(Isaxs, setup, qv_init,
                          phir=phiv_init, mirror=mirror)

    setup.dqv = qv_init[:,1]
    setup.phiv = phiv_init
    setup.qv = np.tile(qv_init[:,0],setup.phiv.shape[0])
    setup.radii = radii

    setup.gproi = np.array([len(x[0]) for x in qroi], dtype=int)
    setup.qroi = qroi
    setup.qv = setup.qv[:len(setup.qroi)]

    xmin = min([x[0].min() for x in setup.qroi])
    ymin = min([x[1].min() for x in setup.qroi])
    xmax = max([x[0].max() for x in setup.qroi])
    ymax = max([x[1].max() for x in setup.qroi])

    qsec = ((xmin, ymin), (xmax, ymax))
    qsec_dim = (xmax - xmin + 1, ymax - ymin + 1)
    setup.qsec = qsec
    setup.qsec_dim = qsec_dim
    setup.qsec_mask = setup.mask[qsec[0][0]:qsec_dim[0]+qsec[0][0], qsec[0][1]:qsec_dim[1]+qsec[0][1]]
    setup.qsec_center = (setup.center[0]-qsec[0][1], setup.center[1]-qsec[0][0])
    setup.qsec_ai = setup.update_ai(setup.qsec_center)

    print('Added the following Q-values [nm-1]:\n{}'.format(setup.qv))

    if plot:
        plotqrois(Isaxs, setup, method=plot, d=d, shade=True, mirror=mirror,
                  **kwargs)



import h5py
import numpy as np

#######################################
#--- get information on data series ---#
#######################################    

def get_attrs_from_dict(obj, meta):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(header, p):
        try:
            attr = p[1](header[p[0]])
        except:
            attr = 0
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        header = obj.get_header(filename)
        for key, value in p.items():
            meta[key][i] = get_attr(header, value)
                
def get_header_h5(*args, **kwargs):
    return 0

def get_attrs_h5(obj, meta,):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(f, p, key):
        if p[0] == obj.h5opt['data']:
            attr = f[p[0]].attrs[key]
        else:
            if p[0]:
                attr = f[p[0]]
                if attr.dtype.kind == 'V':
                    attr = np.mean(attr.value['value']).astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'f':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'u':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'i':
                    attr = attr[()]
            else:
                attr = f[obj.h5opt['data']].shape[0]
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        with h5py.File(filename, 'r', driver=obj.h5opt['driver']) as f:
            for key, value in p.items():
                meta[key][i] = get_attr(f, value, key)

                
'''
AGIPD Methods work in progress ...
'''

# class Pulse:
#
#     def __init__(self,):
#         pass
#
# class Train(Pulse):
#
#     def __init__(self,):
#         self.id = None
#         self.npulses = None
#         self.pulse_ids = None
#
# class Run(Train):
#
#     def __init__(self, runid, rundir, attrdict):
#         self.id = runid
#         self.rundir = rundir
#         self.attrdict = attrdict
#         self.ntrains = None
#         self.train_ids = None
#
#     def get_(self):
#         pass
#
# class MidMetadata(Run):
#
#     def __init__(self, *uid, depth=None, datdir='./',):
#         self.datdir = os.path.abspath(datdir)
#         if depth is None:
#             depth = len(uid)
#         self.depth = depth
#
#
#
# def get_attrs_agipd(obj, meta, depth, uid):
#
#     def init_meta(p):
#         for key, value in p.items():
#             meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
#
#     def get_attr(f, p, key):
#         if p[0] == obj.h5opt['data']:
#             attr = f[p[0]].attrs[key]
#         else:
#             if p[0]:
#                 attr = f[p[0]]
#                 if attr.dtype.kind == 'V':
#                     attr = np.mean(attr.value['value']).astype(np.dtype(p[1]))
#                 elif attr.dtype.kind == 'f':
#                     attr = attr[()]#.astype(np.dtype(p[1]))
#                 elif attr.dtype.kind == 'u':
#                     attr = attr[()]#.astype(np.dtype(p[1]))
#                 elif attr.dtype.kind == 'i':
#                     attr = attr[()]
#             else:
#                 attr = f[obj.h5opt['data']].shape[0]
#         return attr
#
#     f_files = getfiles(obj.datdir, obj.suffix, obj.fastname)
#     s_files = getfiles(obj.datdir, obj.suffix, obj.slowname)
#     print(f_files,s_files)
#
#     p = obj.attributes
#     s = obj.slowdata
#     init_meta(p)
#     for i, m in enumerate(meta.copy()['master']):
#         filename = obj.datdir + m
#         with h5py.File(filename, 'r', driver=obj.h5opt['driver']) as f:
#             for key, value in p.items():
#                 meta[key][i] = get_attr(f, value, key)

def common_mode_from_hist(im, searchoffset=50):                        
    bins = np.arange(-searchoffset,searchoffset) 
    hist, _ = np.histogram(im,bins)
    CM = np.where(hist==hist.max())[0][0] - searchoffset
    return CM # has to be subtracted from the tail



from importlib import import_module
from pyFAI.detectors import Detector


def grab(name, *args, **kwargs):

    try:
        if '.' in name:
            module_name, class_name = name.rsplit('.', 1)
        else:
            module_name = name
            class_name = name.capitalize()

        detector_module = import_module('.' + module_name, package='Xana.detectors')

        detector_class = getattr(detector_module, class_name)

        instance = detector_class(*args, **kwargs)

    except (AttributeError, ModuleNotFoundError):
        raise ImportError('{} is not part of our detector collection!'.format(name))
    else:
        if not issubclass(detector_class, Detector):
            raise ImportError("We currently don't have {}, but you are welcome to send in the request for it!".format(detector_class))

    return instance



from .ReadData import read_data
from .getmeta import get_attrs_from_dict, get_attrs_h5, get_header_h5
from .H5Methods import common_mode_from_hist
import warnings

class Xfmt:
    """Class that defines all methods and attributes to handle
    dfferent data types.
    """

    def __init__(self, fmtstr=''):

        self.fmtstr = fmtstr
        kernel = {}
        if fmtstr == '':
            warnings.warn('Format string is empty. Specify valid data format to load data.')
        else:
            if fmtstr == 'id10_eiger_single_edf':
                kernel = self.__init_id10_eiger_single_edf()
            elif fmtstr == 'pilatus_single_cbf':
                kernel = self.__init_pilatus_single_cbf()
            elif fmtstr == 'p10_eiger_h5':
                kernel = self.__init_p10_eiger_h5()
            elif fmtstr == 'xcs_cspad_h5':
                kernel = self.__init_xcs_cspad_h5()
            elif fmtstr == 'converted_h5':
                kernel = self.__init_converted_h5()
            elif fmtstr == 'lambda_nxs':
                kernel = self.__init_lambda_nxs()
            elif fmtstr == 'id02_eiger_single_edf':
                kernel = self.__init_id02_eiger_single_edf()
            elif fmtstr == 'id02_eiger_multi_edf':
                kernel = self.__init_id02_eiger_multi_edf()
            elif fmtstr == 'spb_agipd':
                kernel = self.__init_spb_agipd()

            self.suffix = None
            self.numfmt = None
            self.masterfmt = None
            self.seriesfmt = None
            self.get_header = None
            self.get_attributes = None
            self.__dict__.update(kernel)

            if 'agipd' not in fmtstr:
                self.load_data_func = read_data

    @staticmethod
    def __init_id10_eiger_single_edf():
        from . import EdfMethods as edf
        kernel = {
            'prefix': 'img_eiger',
            'suffix': '((edf)$|(edf\.gz)$)',
            'numfmt': '\d{4}',
            'masterfmt': '((img_eiger_\d{4}_0000_0000)|(img_0{4})|(zaptime_\d{,5}_eiger1_0{4}))',
            'seriesfmt': '\d{1,5}',
            'get_header': edf.headeredf,
            'get_attributes': get_attrs_from_dict,
            'attributes': {
                't_exposure': ('acq_expo_time', float),
                't_readout': ('ccd_readout_time', float),
                't_latency': ('acq_latency_time', float),
                'nframes': ('acq_nb_frames', int)},
        }
        return kernel

    @staticmethod
    def __init_id02_eiger_single_edf():
        from . import EdfMethods as edf
        kernel = {
            'prefix': '',
            'suffix': '((edf)$|(edf\.gz)$)',
            'numfmt': '((?<=_)\d{4,}(?=\.))',
            'masterfmt': '(.*_\d{5}_00001)',
            'seriesfmt': '\d{5}',
            'get_header': edf.headeredf,
            'get_attributes': get_attrs_from_dict,
            'attributes': {
                't_exposure': ('acq_expo_time', float),
                't_readout': ('ccd_readout_time', float),
                't_latency': ('acq_latency_time', float),
                'nframes': ('acq_nb_frames', int)},
        }
        return kernel

    @staticmethod
    def __init_id02_eiger_multi_edf():
        from . import EdfMethods as edf
        kernel = {
            'prefix': '',
            'suffix': '((edf)$|(edf\.gz)$)',
            'numfmt': '((?<=_)\d{4,}(?=\.))',
            'masterfmt': '(.*_\d{5}_00001)',
            'seriesfmt': '\d{1,5}',
            'get_header': edf.headeredf,
            'get_attributes': get_attrs_from_dict,
            'attributes': {
                't_exposure': ('acq_expo_time', float),
                't_readout': ('ccd_readout_time', float),
                't_latency': ('acq_latency_time', float),
                'nframes': ('acq_nb_frames', int)},
        }
        return kernel

    @staticmethod
    def __init_pilatus_single_cbf():
        from . import CbfMethods as cbf
        kernel = {
            'prefix': 'img_eiger',
            'suffix': '(cbf)$',
            'numfmt': '\d{5}(?=\.)',
            'masterfmt': '.*_\d{5}_00001',
            'seriesfmt': '(?<=_)\d{5}(?=_)',
            'get_header': cbf.headercbf,
            'get_attributes': get_attrs_from_dict,
            'attributes': {
                't_exposure': ('acq_expo_time', float),
                't_readout': ('ccd_readout_time', float),
                't_latency': ('acq_latency_time', float),
                'nframes': ('acq_nb_frames', int)},
        }
        return kernel

    @staticmethod
    def __init_p10_eiger_h5():
        kernel = {
            'prefix': '',
            'suffix': 'h5',
            'numfmt': '\d{5,6}',
            'masterfmt': '.*_master',
            'seriesfmt': '\d{5}((?=_master)|(?=_data))',
            'get_header': get_header_h5,
            'get_attributes': get_attrs_h5,
            'attributes': {
                't_delay': ('/entry/instrument/detector/frame_time', 'float32'),
                't_exposure': ('/entry/instrument/detector/count_time', 'float32'),
                't_readout': ('/entry/instrument/detector/detector_readout_time', 'float32'),
                'nframes': ('/entry/instrument/detector/detectorSpecific/nimages', 'int32')
            },
            'h5opt': {'driver': 'stdio',
                      'data': '/entry/data/',
                      'ExternalLinks': True,
                      'chunk_size': 200,
                      'images_per_file': 2000,
                      },
        }
        return kernel

    @staticmethod
    def __init_xcs_cspad_h5():
        from .ArrangeModules import arrange_cspad_tiles
        from ..Xdrop.DropletizeData import dropletizedata
        kernel = {
            'prefix': '',
            'suffix': 'h5',
            'numfmt': '(?<=-r)\d{4,5}',
            'masterfmt': '.*-r\d{4,5}',
            'seriesfmt': '(?<=-r)\d{4,5}',
            'get_header': get_header_h5,
            'get_attributes': get_attrs_h5,
            'attributes': {
                'rate': ('/Configure:0000/Run:0000/CalibCycle:0000/Epics::EpicsPv/' \
                         'EpicsArch.0:NoDevice.0/Rate/data', 'float32'),
                'pulseLength': ('/Configure:0000/Run:0000/CalibCycle:0000/Epics::EpicsPv/' \
                                'EpicsArch.0:NoDevice.0/Pulse length/data', 'float32'),
                'photonEnergy': ('/Configure:0000/Run:0000/CalibCycle:0000/Epics::EpicsPv/' \
                                 'EpicsArch.0:NoDevice.0/Photon beam energy/data', 'float32'),
                'nframes': (False, 'int32')
            },

            'h5opt': {'driver': 'stdio',
                      'data': '/Configure:0000/Run:0000/CalibCycle:0000/' \
                      'CsPad::ElementV2/XcsEndstation.0:Cspad.0/data',
                      'ExternalLinks': False,
                      'chunk_size': 256,
                      'commonmode': common_mode_from_hist,
                      'arrange_tiles': arrange_cspad_tiles,
                      'dropletize': dropletizedata
                      },
        }
        return kernel

    @staticmethod
    def __init_converted_h5():
        kernel = {
            'prefix': '',
            'suffix': 'h5',
            'numfmt': '(?<=_)(\d{4})',  # '|(_\d{4,}(?=\.)))',
            'masterfmt': '((img_eiger_\d{4}_0000_0000)|(img_0{4})|(zaptime_\d{,5}_eiger1_0{4}))',
            'seriesfmt': '\d{1,5}',
            'get_header': get_header_h5,
            'get_attributes': get_attrs_h5,
            'attributes': {
                't_exposure': ('/data/', 'float32'),
                't_readout': ('/data/', 'float32'),
                't_latency': ('/data/', 'float32'),
                'nframes': ('/data/', 'int32')
            },
            'h5opt': {'driver': 'stdio',
                      'data': '/data/',
                      'ExternalLinks': False,
                      'chunk_size': 200,
                      'images_per_file': 1e6,
                      },
        }
        return kernel

    @staticmethod
    def __init_lambda_nxs():
        kernel = {
            'prefix': '',
            'suffix': 'nxs',
            'numfmt': '(?<=_)\d{5}(?=-)',
            'masterfmt': '.*_\d{5,}-\d{5,}',
            'seriesfmt': '(?<=_)\d{5,}(?=_)',
            'get_header': get_header_h5,
            'get_attributes': get_attrs_h5,
            'attributes': {
                't_exposure': ('/entry/instrument/detector/count_time', 'float32'),
                't_readout': ('/entry/instrument/detector/detector_readout_time', 'float32'),
                't_latency': ('/entry/instrument/detector/collection/delay_time', 'float32'),
                'nframes': ('/entry/instrument/detector/collection/frame_numbers', 'int32')
            },
            'h5opt': {'driver': 'stdio',
                      'data': '/entry/instrument/detector/data',
                      'ExternalLinks': False,
                      'chunk_size': 200,
                      'images_per_file': 1e6,
                      },
        }
        return kernel

    @staticmethod
    def __init_spb_agipd():
        # from .getmeta import get_attrs_agipd
        # from .ArrangeModules import arrange_cspad_tiles
        # from ..Xdrop.DropletizeData import dropletizedata
        kernel = {
            'prefix': '',
            'suffix': 'h5',
            'numfmt': '',
            'masterfmt': '',
            'seriesfmt': '(?<=-R)\d{4}',
            'runfolder': 'r{:04}',
            'fastname': '(?<=AGIPD00-S)\d{5}',
            'slowname': '(?<=DA)\d{2}-S\d{5}',
            'get_attributes': get_attrs_agipd,
            'fastdata': {
                'rate': (''),
                'trainId': ('/INDEX/trainId', 'int32'),
                'pulseId': ('INSTRUMENT/SPB_DET_AGIPD1M-1/DET/0CH0:xtdf/image/pulseId',
                            'int32'),
                'pulseCount': ('INSTRUMENT/SPB_DET_AGIPD1M-1/DET/0CH0:xtdf/header/' \
                               'pulseCount', 'int32')
            },
            'slowdata': {
                'trainId':('/INDEX/trainId', 'int32'),
                'photonEnergy':('CONTROL/ACC_SYS_DOOCS/CTRL/BEAMCONDITIONS/energy/value',
                                 'float32'),
                'photonFluc':('CONTROL/SPB_XTD9_XGM/XGM/DOOCS/pulseEnergy/photonFlux/value',
                              'float32'),
                'nframes': (False, 'int32')
            },

            'h5opt': {'driver': 'stdio',
                      'chunk_size': 256,
                      },
        }
        return kernel
    



import numpy as np
from matplotlib import pyplot as plt
import matplotlib
import itertools
import copy
import pandas as pd

from Decorators import Decorators
from Analist import AnaList
from XParam.parameters import plot_parameters
from Xfit.fitg2 import fitg2
from Xplot.niceplot import niceplot
from misc.resample import resample as resample_func
from XpcsAna.StaticContrast import staticcontrast


class CorrFunc(AnaList):

    def __init__(self, Xana, **kwargs):
        super().__init__(Xana, **kwargs)
        self.fit_result = None
        self.pars = None
        self.corrFunc = None
        self.corrFuncRescaled = None
        self.staticContrast = None
        self.twotime = None
        self.g2plotl = None
        self.nq = np.arange(len(self.Xana.setup['qroi']))
        self.db_id = None

    def __str__(self):
        return 'Corr Func class for g2 displaying.'

    def __repr__(self):
        return 

    def __add__(self, cf2):
        cf3 = CorrFunc(self.Xana)
        cf3.fit_result = self.fit_result.extend(cf2.fit_result)
        cf3.pars = self.pars.extend(cf2.pars)
        cf3.corrFunc = self.corrFunc.extend(cf2.corrFunc)
        cf3.corrFuncRescaled = self.corrFuncRescaled.extend(cf2.corrFuncRescaled)
        cf3.db_id = np.append(self.db_id, cf2.db_id)
        return cf3

    @Decorators.input2list
    def get_g2(self, db_id, merge='append'):
        self.db_id = db_id
        self.corrFunc = []
        if merge == 'merge':
            self.merge_g2(db_id)
        elif merge == 'append':
            for sid in db_id:
                try:
                    d = self.Xana.get_item(sid)
                    self.corrFunc.append((d['corf'], d['dcorf']))
                except KeyError:
                    print('Could not load item {}'.format(sid))
            print('Loaded {} correlation functions.'.format(len(db_id)))
        self.corrFuncRescaled = copy.deepcopy(self.corrFunc)

    @Decorators.init_figure()
    def plot_g2(self, nq=None, err=True, ax=None, nmodes=1, data='rescaled', cmap='jet',
                change_marker=False, color_mode=0, color='b', dofit=True, **kwargs):
        if data == 'original' or self.corrFuncRescaled is None:
            corrFunc = list(self.corrFunc)
        elif data == 'rescaled':
            corrFunc = list(self.corrFuncRescaled)
        else:
            raise ValueError('No usable correlation data defined.')

        if nq is None:
            pass
        elif type(nq) == int:
            self.nq = np.arange(nq)
        else:
            self.nq = nq

        if color_mode < 2:
            if color_mode == 0:
                color_multiplier, color_repeater = self.nq.size, len(corrFunc)
            elif color_mode == 1:
                color_multiplier, color_repeater = self.nq.size * \
                    len(corrFunc), 1
            self.update_colors(cmap, color_multiplier, color_repeater)
        elif color_mode == 2:
            self.colors = [color]*len(self.nq)*len(corrFunc)

        self.update_markers(len(corrFunc), change_marker)

        self.g2plotl = [[]] * len(corrFunc)
        self.pars = [[]] * len(corrFunc)
        self.fit_result = [[[] for i in range(self.nq.size)]] * len(corrFunc)

        ci = 0
        for j, (cfi, dcfi) in enumerate(corrFunc):
            rates = np.zeros((self.nq.size, 3*nmodes+3, 2))
            ti = cfi[1:, 0]
            for i, qi in enumerate(self.nq):
                if i == 0:
                    cf_id = self.db_id[j]
                else:
                    cf_id = None
                res = fitg2(ti, cfi[1:, qi+1], err=dcfi[1:, qi+1],
                            qv=cfi[0,1+qi],
                            ax=ax, color=self.colors[ci],
                            marker=self.markers[j % len(self.markers)], cf_id=cf_id,
                            modes=nmodes, dofit=dofit, **kwargs)
                self.fit_result[j][i] = res[2:4]
                self.g2plotl[j].append(
                    list(itertools.chain.from_iterable(res[4])))
                if dofit:
                    if i == 0:
                        db_tmp = self.init_pars(list(res[2].params.keys()))
                    entry = [cfi[0, qi+1], *res[0].flatten(), *res[1]]
                    db_tmp.loc[i] = entry
                else:
                    db_tmp = 0
                ci += 1
            self.pars[j] = db_tmp

    @staticmethod
    def init_pars(names):
        names = [names[i//2] if (i+1) % 2 else 'd'+names[i//2]
                 for i in range(len(names)*2)]
        names.insert(0, 'q')
        names.extend(['chisqr', 'redchi', 'bic', 'aic'])
        return pd.DataFrame(columns=names)

    def reset_rescaled(self):
        self.corrFuncRescaled = copy.deepcopy(self.corrFunc)

    def rescale(self, index=None, normby='average', norm_baseline=True,
                   norm_contrast=False, nq=None, baseline=1., contrast=None,
                   interval=(1, -1), weighted=True):

        def rescale(y, mn, mx, rng=(0, 1)):
            p = (rng[1]-rng[0])/(mx-mn)
            return p * (y - mn) + rng[0], p

        def normFunc(corrFunc, pars):
            norm_b = np.min(corrFunc[0][1:, nq+1], axis=0)
            norm_c = np.max(corrFunc[0][1:, nq+1], axis=0)
            if normby == 'fit':
                for iq in range(nq.size):
                    norm_b[iq] = pars.loc[iq, 'a']
                    if norm_contrast:
                        norm_c[iq] = pars.loc[iq, 'beta'] + pars.loc[iq, 'a']
            elif normby == 'average':
                for iq in range(nq.size):
                    if weighted:
                        weights = 1/corrFunc[1][interval[1]:, nq[iq]+1]**2
                    else:
                        weights = None
                    norm_b[iq] = np.ma.average(
                        corrFunc[0][interval[1]:, nq[iq]+1], weights=weights)
                    if norm_contrast:
                        if weighted:
                            weights = 1 / \
                                corrFunc[1][1:max([interval[0],1])+1, nq[iq]+1]**2
                        else:
                            weights = None
                        norm_c[iq] = np.ma.average(
                            corrFunc[0][1:max([interval[0],1])+1, nq[iq]+1],
                            weights=weights)

            if contrast is None:
                initial_contrast = norm_c - norm_b
            else:
                initial_contrast = contrast
            corrFunc[0][1:, nq+1], p = rescale(
                corrFunc[0][1:, nq+1], norm_b, norm_c,
                (baseline, initial_contrast + baseline))
            corrFunc[1][1:, nq+1] *= p

        if self.pars is None:
            self.pars = [None] * len(self.corrFunc)
        if nq is None:
            nq = self.nq
        if index is None:
            index = slice(len(self.corrFuncRescaled))
        else:
            index = slice(index, index+1, 1)

        for corrFunc, pars in zip(self.corrFuncRescaled[index], self.pars[index]):
            normFunc(corrFunc, pars)

    def rescale_user(self, offset=0., nq=None):

        if nq is None:
            nq = self.nq

        for corrFunc, o in zip(self.corrFuncRescaled, offset):
            corrFunc[0][1:, nq+1] += o

    def merge_g2(self, in_list, limit=0.005):
        t_exp = np.zeros(len(in_list))
        nframes = t_exp.copy()
        for ii, i in enumerate(in_list):
            t_exp[ii] = self.Xana.db.loc[i]['t_exposure']
            nframes[ii] = self.Xana.db.loc[i]['nframes']

        ind = np.argsort(nframes)[::-1]
        t_exp = t_exp[ind]
        nframes = nframes[ind]
        in_list = np.array(in_list)[ind]

#        if (np.unique(t_exp, return_inverse=True)[1] == np.unique(nframes, return_inverse=True)[1]).any():
        uq_et, uq_inv, uq_cnt,  = np.unique(
            t_exp, return_inverse=1, return_counts=1)
#        else:
 #           raise AssertionError('Cannot merge correlation functions.')

        qv = self.Xana.setup['qv']
        counter = np.zeros(uq_et.size, dtype=np.int32)
        for i, cnti in enumerate(uq_cnt):
            for j in range(cnti):
                ind = np.where(uq_inv == i)[0][j]
                counter[i] += nframes[ind]
                item = in_list[ind]
                try:
                    d = self.Xana.get_item(item)
                    if j == 0:
                        t = d['corf'][1:, 0]
                        cf = np.zeros((cnti, t.size, qv.size))
                        dcf = np.zeros_like(cf)
                    cf[j, :, :qv.size] = d['corf'][1:, 1:qv.size+1]
                    dcf[j, :, :qv.size] = d['dcorf'][1:, 1:qv.size+1]
                except ValueError as v:
                    print('Tried loading database entry: ', item)
                    raise ValueError(v)

            cf = np.ma.masked_array(cf, mask=((cf < limit) | np.isnan(cf) | (dcf <= 0)))
            dcf = np.ma.masked_array(dcf, mask=cf.mask)
            dcf = 1/(dcf**2)# + np.ma.var(cf))
            cf, dcf = np.ma.average(cf, weights=dcf, returned=1, axis=0)
            cf = np.hstack((t[:, None], cf.filled(np.nan)))
            dcf[dcf > 0] = np.sqrt(1/dcf[dcf > 0])
            dcf = np.hstack((t[:, None], dcf.filled(np.nan)))
            self.corrFunc.append((np.vstack((np.append(0, qv), cf)),
                                  np.vstack((np.append(0, qv), dcf))))

        tmp = "Merged g2 functions: "
        print('{:<22}{} (exposure times)'.format(tmp, np.round(uq_et, 6)))
        print('{:<22}{} (number of correlation functions)'.format('', uq_cnt))
        print('{:<22}{} (total number of images)'.format('', counter))

    def merge_g2list(self, resample=False, cutoff=-1, **kwargs):
        for ii, cf_master in enumerate([self.corrFunc, self.corrFuncRescaled]):
            if cf_master is not None:
                for i, cf in enumerate(cf_master):
                    if i == 0:
                        cf_tmp = cf[0][:cutoff]
                        dcf_tmp = cf[1][:cutoff]
                    else:
                        cf_tmp = np.vstack((cf_tmp, cf[0][1:cutoff]))
                        dcf_tmp = np.vstack((dcf_tmp, cf[1][1:cutoff]))

                if resample:
                    new_t, new_cf, new_dcf = resample_func(cf_tmp[1:, 0], cf_tmp[1:, 1:], dcf_tmp[1:, 1:],
                                                           resample, **kwargs)
                    cf_tmp = np.hstack((new_t[:, None], new_cf))
                    cf_tmp = np.vstack((cf_master[0][0][0, :], cf_tmp))
                    dcf_tmp = np.hstack((new_t[:, None], new_dcf))
                    dcf_tmp = np.vstack((cf_master[0][0][0, :], dcf_tmp))

                if ii == 0:
                    self.corrFunc = [(cf_tmp, dcf_tmp), ]
                elif ii == 1:
                    self.corrFuncRescaled = [(cf_tmp, dcf_tmp), ]

    def plot_parameters(self, plot, ax=None, change_axes=True,
                        cindoff=0, **kwargs):
        """Plot Fit parameter (decay rates, kww exponent, etc.)
        """
        npl = len(plot)
        npars = len(self.pars)

        if ax is None and change_axes:
            n = int(np.ceil(npl/2))
            fig, ax = plt.subplots(
                n, int(npl > 1)+1, figsize=(5+4*(npl > 1), 4*n))
        elif ax is None and not change_axes:
            fig, ax = plt.subplots(1, 1, figsize=(5, 4))

        if isinstance(ax, np.ndarray):
            ax = ax.flatten()
        else:
            ax = [ax,]

        mm = max([len(x) if isinstance(x, (list,tuple)) else 1    
                  for x in kwargs.get('modes',[1])])
            
        if change_axes:
            ax_idx = np.arange(len(ax))
            c_idx = np.repeat(np.arange(mm*npars), npl)
        else:
            ax_idx = np.zeros(npl, dtype=np.int16)
            c_idx = np.arange(npl*npars)

        for ipar, pars in enumerate(self.pars):
            for i, p in enumerate(plot):
                kwargsl = {key: value[i] if type(value) == tuple else value \
                           for (key, value) in kwargs.items()}
                plot_parameters(pars, p, ax=ax[ax_idx[i]], ci=c_idx[ipar:ipar+mm],
                                 **kwargsl)
        plt.tight_layout()
        return ax

    @Decorators.input2list
    def get_twotime(self, db_id, twotime_par=None):
        """Receive two-time correlation functions from database
        """
        self.twotime = 0.
        for i, sid in enumerate(db_id):
            d = self.Xana.get_item(sid)
            if twotime_par is None:
                twotime_par = d['twotime_par']
            self.twotime += d['twotime_corf'][twotime_par]
        self.twotime /= i + 1

    @Decorators.init_figure()
    @Decorators.input2list
    def plot_twotime(self, db_id, clim=(None, None), ax=None, interpolation='gaussian'):
        """Plot two-time correlation functions read from database
        """
        self.get_twotime(db_id)

        vmin, vmax = clim
        corfd = self.Xana.get_item(db_id[0])
        ax.set_title(
            r'q = {:.2g}$\mathrm{{nm}}^{{-1}}$'.format(corfd['qv'][corfd['twotime_par']]))
        tt = corfd['twotime_xy']
        im = ax.imshow(self.twotime, cmap=plt.get_cmap('jet'), origin='lower',
                       interpolation=interpolation, extent=[
                           tt[0], tt[-1], tt[0], tt[-1]],
                       vmin=vmin, vmax=vmax)
        ax.set_xlabel(r'$t_1$ [s]',)
        ax.set_ylabel(r'$t_2$ [s]',)
        cl = plt.colorbar(im, ax=ax)
        cl.ax.set_ylabel('correlation', fontsize=12)
        niceplot(ax, autoscale=False, grid=False)

    @Decorators.init_figure()
    @Decorators.input2list
    def plot_trace(self, db_id, log='', ax=None):
        axtop = ax.twiny()

        ci = 0
        for sid in db_id:
            corfd = self.Xana.get_item(sid)
            trace = corfd['trace']
            framen = np.arange(trace.shape[0])
            time0 = corfd['twotime_xy'][0]
            time1 = corfd['twotime_xy'][-1]
            time = np.linspace(time0, time1, framen.size)
            for i, iq in enumerate(self.nq):
                ax.plot(
                    time, trace[:, iq], 'o', color=self.colors[ci], markersize=2, label=str(i))
                axtop.plot(
                    framen, trace[:, iq], 'o', color=self.colors[ci], markersize=2, label=str(i))
                ci += 1

        ax.set_ylabel('photons per pixel')
        ax.set_xlabel('time in [s]')
        niceplot(ax, autoscale=0)
        axtop.set_xlabel('frame number')
        niceplot(axtop, autoscale=False, grid=False)
        if 'x' in log:
            ax.set_xscale('log')
        if 'y' in log:
            ax.set_yscale('log')

        plt.legend()

        # ax.set_title('trace', fontweight='bold', fontsize=14, y=1.14)

    def get_static_contrast(self, *args, **kwargs):
        self.staticContrast = staticcontrast(self, *args, **kwargs)

    def g2_totxt(self, savname):
        return 0



import numpy as np
from matplotlib import pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
import copy
import pandas as pd
import collections

from ..Decorators import Decorators
from ..Analist import AnaList
from ..XParam.parameters import plot_parameters
from ..Xfit.fitg2global import G2
from ..Xplot.niceplot import niceplot
from ..misc.resample import resample as resample_func
from .StaticContrast import staticcontrast


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
        updating only top-level keys, dict_merge recurses down into dicts nested
        to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
        ``dct``.
        :param dct: dict onto which the merge is executed
        :param merge_dct: dct merged into dct
        :return: None
    """
    for k, v in merge_dct.items():
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], collections.Mapping):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


class CorrFunc(AnaList):

    def __init__(self, Xana, **kwargs):
        super().__init__(Xana, **kwargs)
        self.fit_result = [None]
        self.pars = [None]
        self.corrFunc = [None]
        self.corrFuncRescaled = [None]
        self.staticContrast = [None]
        self.corrFuncChi2 = [None]
        self.twotime = [None]
        self.g2plotl = [None]
        self.nq = np.arange(len(self.Xana.setup.qroi))
        self.db_id = []
        self.ncfs = 0
        self.fit_config = []

    def __str__(self):
        return 'Corr Func class for g2 displaying.'

    def __getstate__(self):
        d = dict(vars(self))
        # d['twotime'] = None
        d['g2plotl'] = None
        return d

    def __setstate__(self, d):
        # deal with old type of db_id and fit_config properties
        if isinstance(d['db_id'], np.ndarray):
            d['db_id'] = [d['db_id'],]
        if isinstance(d['fit_config'], dict):
            d['fit_config'] = [d['fit_config'],]
        self.__dict__.update(d)

    def __add__(self, cf2):
        cf3 = CorrFunc(self.Xana)
        extend_items = [
            'corrFunc',
            'corrFuncRescaled',
            'db_id',
            'pars',
            'fit_result',
            'fit_config',
            'corrFuncChi2',
        ]

        for item in extend_items:
            value = copy.deepcopy(getattr(self, item))
            value.extend(getattr(cf2, item))
            setattr(cf3, item, value)

        return cf3

    @Decorators.input2list
    def get_g2(self, db_id, merge='append', **kwargs):
        self.corrFunc = []
        if merge == 'merge':
            self.merge_g2(db_id, **kwargs)
        elif merge == 'append':
            for sid in db_id:
                self.db_id.append([sid])
                try:
                    d = self.Xana.get_item(sid)
                    self.corrFunc.append((np.ma.masked_invalid(d['corf']),
                                          np.ma.masked_invalid(d['dcorf'])))
                except KeyError:
                    print('Could not load item {}'.format(sid))
            print('Loaded {} correlation functions.'.format(len(db_id)))
        self.corrFuncRescaled = copy.deepcopy(self.corrFunc)

    @Decorators.init_figure()
    def plot_g2(self, nq=None, ax=None, data='rescaled', cmap='magma',
                change_marker=False, color_mode=0, color='b', dofit=False,
                index=None, exclude=None, add_colorbar=False, cb_kws={},
                **kwargs):

        fit_keys = ['nmodes', 'mode', 'fix', 'init', 'fitglobal', 'lmfit_pars', 'fitqdep']
        fit_kwargs = {k:i for k,i in kwargs.items() if k in fit_keys}

        if data == 'original' or self.corrFuncRescaled is None:
            corrFunc = list(self.corrFunc)
        elif data == 'rescaled':
            corrFunc = list(self.corrFuncRescaled)
        else:
            raise ValueError('No usable correlation data defined.')

        ind_cfs = np.arange(len(corrFunc))
        ind_pars = ind_cfs.copy()
        if index is not None:
            if isinstance(index, (int, np.integer)):
                s = [index]
            elif isinstance(index, (list, tuple)):
                s = np.int32(index)
            elif isinstance(index, np.ndarray):
                s = index
            else:
                raise ValueError(f'Index of type {type(index)} not supported. User int or list.')
            corrFunc = [corrFunc[si] for si in s]
            ind_pars = ind_pars[s]
            ind_cfs = np.arange(len(corrFunc))
        elif index is None and exclude is not None:
            ind_cfs = np.delete(ind_cfs, exclude)
            ind_pars = np.delete(ind_pars, exclude)

        ncf = ind_cfs.size

        if nq is None:
            pass
        elif type(nq) == int:
            self.nq = np.arange(nq)
        else:
            self.nq = nq
        self.qv = self.Xana.setup.qv[self.nq]

        if color_mode in [0, 1]:
            if color_mode == 0:
                color_multiplier, color_repeater = self.nq.size, ncf
            elif color_mode == 1:
                color_multiplier, color_repeater = self.nq.size * \
                    ncf, 1
            self.update_colors(cmap, color_multiplier, color_repeater)
        elif color_mode == 2:
            self.colors = [color]*len(self.nq)*ncf
        elif color_mode == 'from func' or color_mode == 3:
            self.colors = [cmap(x) for x in self.qv]
        elif color_mode == 'from vec' or color_mode == 4:
            self.colors = [cmap(x) for x in self.qv]
        self.update_markers(ncf, change_marker)

        if self.fit_config is None:
            self.fit_config = [dict(fit_kwargs),] * ncf
        elif isinstance(self.fit_config, list):
            if len(self.fit_config) == 0:
                self.fit_config = [dict(fit_kwargs),] * ncf
            for d in self.fit_config:
                dict_merge(d, fit_kwargs)

        if dofit:
            self.fit_result = [[]] * ncf

        if dofit or self.pars is None or (len(self.pars)<=max(ind_pars)) :
            self.pars = [[]] * np.max([ncf, np.max(ind_pars)+1])

        if len(self.pars) > len(self.db_id):
            tmp = [np.where(np.diff(x)<0)[0] for x in self.db_id]
            tmp = [np.split(x, y+1) for x,y in zip(self.db_id,tmp)]
            self.db_id = [item for sublist in tmp for item in sublist]

        ci = 0
        for j, (ipar,icf) in enumerate(zip(ind_pars,ind_cfs)):
            cfi, dcfi = corrFunc[icf]
            g2 = G2(cfi, self.nq, dcfi)
            if dofit:
                res = g2.fit(**self.fit_config[j])
                self.pars[j] = res[0]
                self.fit_result[j] = res[1]
                print(f'Fit successful: {res[1][0][0].errorbars}')
            if 'doplot' in kwargs:
                g2.plot(marker=self.markers[j % len(self.markers)],
                        ax=ax,
                        colors=self.colors[ci*len(self.nq):(ci+1)*len(self.nq)],
                        data_label='id {}'.format(self.db_id[icf][0]), pars=self.pars[ipar],
                        **kwargs)
                ci += 1

                handles = ax.get_lines()[1::2]
                if 'legi' in kwargs['doplot']:
                    labels = [f'index: {si}' for si in s]
                elif 'legu' in kwargs['doplot']:
                    labels = kwargs.pop('user_labels', [])
                    if len(labels) == 0:
                        print('no user labels found')
                else:
                    labels = []
                handles = handles[:len(labels)]
                if len(handles):
                    ax.legend(handles, labels, loc=0)

        if add_colorbar:
            self.add_colorbar(ax, self.qv, cmap=self.colors, discrete=True, **cb_kws)

    @staticmethod
    def add_colorbar(ax, vec, label=None, cmap='magma', discrete=False, tick_step=1,
                     qscale=0, location='right', show_offset=False, **kwargs):

        ncolors = len(vec)
        tick_indices = np.arange(0, len(vec), tick_step)
        vec = np.array(vec)[tick_indices]
        vec *= 10**qscale
        # divider = make_axes_locatable(ax)
        # cax = divider.append_axes('right', size='5%', pad=0.05)

        change_ticks = False
        if discrete:
            norm = mpl.colors.NoNorm()
            change_ticks = True
            if isinstance(cmap, str):
                cmap = plt.get_cmap(cmap, ncolors)
            elif isinstance(cmap, (list, np.ndarray)):
                cmap = mpl.colors.ListedColormap(cmap)
        else:
            cmap = plt.get_cmap(cmap)
            norm = mpl.colors.Normalize(vmin=vec.min(), vmax=vec.max())


        if location == 'right':
            orientation = 'vertical'
        elif location == 'top':
            orientation = 'horizontal'

        cax = mpl.colorbar.make_axes(ax, location=location, **kwargs)[0]
        cb = mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap,
                                       orientation=orientation)

        # set up color bar ticks
        if location == 'top':
            cax.xaxis.set_ticks_position('top')
            cax.xaxis.set_label_position('top')

        cb.set_label(label)
        if change_ticks:
            cb.set_ticks(tick_indices)
            cb.set_ticklabels(list(map(lambda x: "$%.{}f$".format(3-qscale) % x, vec)))
            if qscale and show_offset:
                cb.ax.text(1.,1.04, r'$\times 10^{{-{}}}$'.format(qscale),
                            transform=cb.ax.transAxes)
        cb.ax.invert_yaxis()
        cb.ax.set_in_layout(True)


    @staticmethod
    def init_pars(names, entry0):
        names = [names[i//2] if (i+1) % 2 else 'd'+names[i//2]
                 for i in range(len(names)*2)]
        names.insert(0, entry0)
        names.extend(['chisqr', 'redchi', 'bic', 'aic'])
        return pd.DataFrame(columns=names)

    def reset_rescaled(self):
        self.corrFuncRescaled = copy.deepcopy(self.corrFunc)

    def rescale(self, index=None, normby='average', norm_baseline=True,
                   norm_contrast=False, nq=None, baseline=1., contrast=1.,
                   interval=(1, -1), weighted=False):

        def rescale(y, mn, mx, rng=(0, 1)):
            p = (rng[1]-rng[0])/(mx-mn)
            return p * (y - mn) + rng[0], p

        def normFunc(corrFunc, pars):
            norm_b = np.nanmin(corrFunc[0][1:, nq+1], axis=0)
            norm_c = np.nanmax(corrFunc[0][1:, nq+1], axis=0)
            if normby == 'fit':
                for iq in range(nq.size):
                    norm_b[iq] = pars.loc[iq, 'a']
                    if norm_contrast:
                        norm_c[iq] = pars.loc[iq, 'beta'] + pars.loc[iq, 'a']
            elif normby == 'average':
                for iq in range(nq.size):
                    if weighted:
                        weights = 1/corrFunc[1][interval[1]:, nq[iq]+1]**2
                    else:
                        weights = None
                    norm_b[iq] = np.ma.average(
                        corrFunc[0][interval[1]:, nq[iq]+1], weights=weights)
                    if norm_contrast:
                        if weighted:
                            weights = 1 / \
                                corrFunc[1][1:max([interval[0],1])+1, nq[iq]+1]**2
                        else:
                            weights = None
                        norm_c[iq] = np.ma.average(
                            corrFunc[0][1:max([interval[0],1])+1, nq[iq]+1], weights=weights)

            if norm_contrast is False:
                initial_contrast = norm_c - norm_b
            else:
                initial_contrast = contrast
            corrFunc[0][1:, nq+1], p = rescale(corrFunc[0][1:, nq+1], norm_b, norm_c,
                                               (baseline, initial_contrast + baseline))
            corrFunc[1][1:, nq+1] *= p

        if self.pars is None:
            self.pars = [None] * len(self.corrFunc)
        if nq is None:
            nq = self.nq
        if index is None:
            index = slice(len(self.corrFuncRescaled))
        else:
            index = slice(index, index+1, 1)

        for corrFunc, pars in zip(self.corrFuncRescaled[index], self.pars[index]):
            normFunc(corrFunc, pars)

    def rescale_user(self, ind=None, offset=None, factor=None,  nq=None):

        if nq is None:
            nq = self.nq

        if offset is None:
            offset = np.zeros(len(self.corrFuncRescaled))

        if factor is None:
            factor = np.ones(len(self.corrFuncRescaled))

        for corrFunc, o, f in zip(self.corrFuncRescaled, offset, factor):
            corrFunc[0][1:, nq+1] *= f
            corrFunc[0][1:, nq+1] -= f - 1
            corrFunc[1][1:, nq+1] *= f
            corrFunc[0][1:, nq+1] += o

    def merge_g2(self, in_list, limit=0., chi2sig=3, cutoff=0):
        self.corrFuncChi2 = []

        t_exp = np.zeros(len(in_list))
        nframes = t_exp.copy()
        for ii, i in enumerate(in_list):
            t_exp[ii] = self.Xana.db.loc[i]['t_exposure']
            nframes[ii] = self.Xana.db.loc[i]['nframes']

        ind = np.argsort(t_exp)[::-1]
        t_exp = t_exp[ind]
        nframes = nframes[ind]
        in_list = np.array(in_list)[ind]

        uq_et, uq_inv, uq_cnt,  = np.unique(
            t_exp, return_inverse=True, return_counts=True)

        counter = np.zeros(uq_et.size, dtype=np.int32)
        for i, cnti in enumerate(uq_cnt):
            indall = np.where(uq_inv == i)[0]
            self.db_id.append(indall)
            for j, ind in enumerate(indall):
                counter[i] += nframes[ind]
                item = in_list[ind]
                try:
                    d = self.Xana.get_item(item)
                    if j == 0:
                        qv = d['corf'][0,1:]
                        t = d['corf'][1:, 0]
                        cf = np.zeros((cnti, t.size, qv.size))
                        dcf = np.zeros_like(cf)

                    cft = d['corf'][1:-(cutoff+1), 1:qv.size+1]

                    if cft.shape[0] > t.size:
                        t = d['corf'][1:, 0]
                        pad = np.zeros((cnti, t.size-cf.shape[1], qv.size))
                        cf = np.concatenate((cf, pad), axis=1)
                        dcf = np.concatenate((dcf, pad), axis=1)

                    cf[j, :cft.shape[0], :qv.size] = cft
                    dcf[j, :cft.shape[0], :qv.size] = d['dcorf'][1:-(cutoff+1), 1:qv.size+1]
                except ValueError as v:
                    print('Tried loading database entry: ', item)
                    raise ValueError(v)

            cf = np.ma.masked_array(cf, mask=np.isnan(cf))
            cf = np.ma.masked_less_equal(cf.filled(-1), limit, copy=False)
            dcf = np.ma.masked_array(dcf, mask=np.isnan(dcf))
            dcf = np.ma.masked_where(dcf.filled(-1)<=0, dcf, copy=False)
            cf = np.ma.masked_where(dcf.mask, cf, copy=False)
            dcf = np.ma.masked_where(cf.mask, dcf, copy=False)

            cfm, dcfm = np.ma.average(cf, weights=1/dcf**2, returned=1, axis=0)

            chi2arr = np.ma.sum((cf - cfm)**2 / cfm**2, 1)
            chi2arr /= cf.shape[1] - 1
            chi2arr = chi2arr[:,0]  # np.max(chi2arr, -1)

            chi2cond = chi2arr > (chi2arr.mean() + chi2sig * chi2arr.std())
            chi2ret = (in_list[indall[chi2cond]], chi2arr.compressed())
            self.corrFuncChi2.append(chi2ret)

            cfm = np.ma.hstack((t[:, None], cfm))
            dcfm = np.ma.sqrt(1/dcfm)
            dcfm = np.ma.hstack((t[:, None], dcfm))

            self.corrFunc.append((np.ma.vstack((np.append(0, qv), cfm)),
                                  np.ma.vstack((np.append(0, qv), dcfm))))

        tmp = "Merged g2 functions: "
        print('{:<22}{} (exposure times)'.format(tmp, np.round(uq_et, 6)))
        print('{:<22}{} (number of correlation functions)'.format('', uq_cnt))
        print('{:<22}{} (total number of images)'.format('', counter))

    def merge_g2list(self, resample=False, cutoff=-1, **kwargs):
        self.db_id = [np.hstack(self.db_id), ]
        for ii, cf_master in enumerate([self.corrFunc, self.corrFuncRescaled]):
            if cf_master is not None:
                for i, cf in enumerate(cf_master):
                    if i == 0:
                        cf_tmp = cf[0][:cutoff]
                        dcf_tmp = cf[1][:cutoff]
                    else:
                        cf_tmp = np.vstack((cf_tmp, cf[0][1:cutoff]))
                        dcf_tmp = np.vstack((dcf_tmp, cf[1][1:cutoff]))

                if resample:
                    new_t, new_cf, new_dcf = resample_func(cf_tmp[1:, 0], cf_tmp[1:, 1:],
                                                           dcf_tmp[1:, 1:], resample,
                                                           **kwargs)
                    cf_tmp = np.hstack((new_t[:, None], new_cf))
                    cf_tmp = np.vstack((cf_master[0][0][0, :], cf_tmp))
                    dcf_tmp = np.hstack((new_t[:, None], new_dcf))
                    dcf_tmp = np.vstack((cf_master[0][0][0, :], dcf_tmp))

                indsort = np.argsort(cf_tmp[1:,0])
                cf_tmp[1:] = cf_tmp[indsort+1]
                dcf_tmp[1:] = dcf_tmp[indsort+1]
                if ii == 0:
                    self.corrFunc = [(cf_tmp, dcf_tmp), ]
                elif ii == 1:
                    self.corrFuncRescaled = [(cf_tmp, dcf_tmp), ]

    def plot_parameters(self, plot, cmap='tab10', ax=None, change_axes=True,
                        cindoff=0, extpar_name='extpar', extpar_vec=None,
                        color_discrete=True, exclude=None, include=None, **kwargs):
        """Plot Fit parameter (decay rates, kww exponent, etc.)
        """
        npl = len(plot)
        npars = len(self.pars)

        ind_pars = np.arange(npars)
        if exclude is not None:
            ind_pars = np.delete(ind_pars, exclude)
        elif include is not None:
            ind_pars = np.array(include)
        npars = ind_pars.size

        if ax is None and change_axes:
            n = int(np.ceil(npl/2))
            fig, ax = plt.subplots(
                n, int(npl > 1)+1, figsize=(5+4*(npl > 1), 4*n))
        elif ax is None and not change_axes:
            fig, ax = plt.subplots(1, 1, figsize=(5, 4))

        if isinstance(ax, np.ndarray):
            ax = ax.flatten()
        else:
            ax = [ax,]

        modes = kwargs.get('modes', 1)
        if isinstance(modes, (list, tuple)):
            mm = max([len(x) if isinstance(x, (list,tuple)) else 1 for x in modes])
        else:
            mm = 1

        if change_axes:
            ax_idx = np.arange(len(ax))
            c_idx = np.repeat(np.arange(npars), len(ax_idx))
        else:
            ax_idx = np.zeros(npl, dtype=np.int16)
            c_idx = np.arange(npars)

        if color_discrete:
            cmap = plt.get_cmap(cmap, npars+1)
        else:
            cmap = plt.get_cmap(cmap)

        self.pars2 = [[]] * npl
        self.fit_result2 = [[[] for i in range(npars)]] * npl

        if extpar_vec is None:
            extpar_vec = np.zeros(npars)
        else:
            assert len(extpar_vec) == npars

        for i, p in enumerate(plot):
            pars_initialized = False
            db_tmp = 0
            for ipar, ind_par in enumerate(ind_pars):
                pars = self.pars[ind_par]
                kwargsl = {key: value[i] if isinstance(value, (tuple, list)) else value \
                           for (key, value) in kwargs.items()}
                res = plot_parameters(pars, p, ax=ax[ax_idx[i]], ci=c_idx[ipar], cmap=cmap,
                                      **kwargsl)
                self.fit_result2[i][ipar] = res[2:4]
                if not isinstance(res, np.ndarray):
                    if not pars_initialized:
                        db_tmp = self.init_pars(list(res[2].params.keys()), extpar_name)
                        pars_initialized = True
                    entry = [extpar_vec[ipar], *res[0].flatten(), *res[1]]
                    db_tmp.loc[ipar] = entry
            self.pars2[i] = db_tmp

        plt.tight_layout()

    @Decorators.input2list
    def get_twotime(self, db_id, twotime_par=None):
        """Receive two-time correlation functions from database
        """
        self.twotime = 0.
        i = 0
        for sid in db_id:
            d = self.Xana.get_item(sid)
            if twotime_par is None:
                twotime_par = d['twotime_par']
            try:
                self.twotime += d['twotime_corf'][twotime_par]
                i += 1
            except ValueError as e:
                print('Could not average %d error message was\n\t%s' % (int(sid), e))
        self.twotime /= i


    @Decorators.init_figure()
    @Decorators.input2list
    def plot_twotime(self, db_id, clim=(None, None), ax=None, interpolation='gaussian', \
                     twotime_par=None):
        """Plot two-time correlation functions read from database
        """
        self.get_twotime(db_id, twotime_par)

        vmin, vmax = clim
        corfd = self.Xana.get_item(db_id[0])
        ax.set_title(
            r'q = {:.2g}$\mathrm{{nm}}^{{-1}}$'.format(corfd['qv'][corfd['twotime_par']]))
        tt = corfd['twotime_xy']
        im = ax.imshow(self.twotime, cmap=plt.get_cmap('magma'), origin='lower',
                       interpolation=interpolation, extent=[
                           tt[0], tt[-1], tt[0], tt[-1]],
                       vmin=vmin, vmax=vmax)
        ax.set_xlabel(r'$t_1$ [s]',)
        ax.set_ylabel(r'$t_2$ [s]',)
        cl = plt.colorbar(im, ax=ax)
        cl.ax.set_ylabel('correlation', fontsize=12)
        # niceplot(ax, autoscale=False, grid=False)

    @Decorators.init_figure()
    @Decorators.input2list
    def plot_trace(self, db_id, log='', ax=None, cmap='Set1'):
        axtop = ax.twiny()

        ci = 0
        cmap = plt.get_cmap(cmap)
        for sid in db_id:
            corfd = self.Xana.get_item(sid)
            trace = corfd['trace']
            framen = np.arange(trace.shape[0])
            time0 = corfd['twotime_xy'][0]
            time1 = corfd['twotime_xy'][-1]
            time = np.linspace(time0, time1, framen.size)
            for i, iq in enumerate(self.nq):
                ax.plot(
                    time, trace[:, iq], 'o-', color=cmap(ci), markersize=2, label=str(i))
                axtop.plot(
                    framen, trace[:, iq], 'o-', color=cmap(ci), markersize=2, label=str(i))
                ci += 1

        ax.set_ylabel('photons per pixel')
        ax.set_xlabel('time in [s]')
        # niceplot(ax, autoscale=0)
        axtop.set_xlabel('frame number')
        # niceplot(axtop, autoscale=False, grid=False)
        if 'x' in log:
            ax.set_xscale('log')
        if 'y' in log:
            ax.set_yscale('log')

        plt.legend()

        # ax.set_title('trace', fontweight='bold', fontsize=14, y=1.14)

    def get_static_contrast(self, *args, **kwargs):
        self.staticContrast = staticcontrast(self, *args, **kwargs)

    def g2_totxt(self, savname):
        return 0



import numpy as np
from matplotlib import pyplot as plt
import matplotlib
import pandas as pd

from ..Decorators import Decorators
from ..Xplot.niceplot import niceplot

@Decorators.init_figure((2,2),(9,8))
def staticcontrast(obj, nq=None, ax=None, data='original', cmap='magma',
                change_marker=False, color_mode=1, t_max=np.inf, trace=True,
                markersize=3, **kwargs):

    if data == 'original':
        corrFunc = list(obj.corrFunc)
    elif data == 'rescaled':
        corrFunc = list(obj.corrFuncRescaled)
    else:
        raise ValueError('No usable correlation data defined.')

    if nq is None:
        pass
    elif type(nq) == int:
        obj.nq = np.arange(nq)
    else:
        obj.nq = nq

    if color_mode == 0:
        color_multiplier, color_repeater = obj.nq.size, len(corrFunc)
    elif color_mode == 1:
        color_multiplier, color_repeater = obj.nq.size*len(corrFunc), 1

    obj.update_colors(cmap, color_multiplier, color_repeater)
    obj.update_markers( len(corrFunc), change_marker)

    staticContrast = [[]] * len(corrFunc)

    tr = np.zeros((len(obj.db_id), len(obj.nq), 2))
    for i,ii in enumerate(obj.db_id):
        tmp = obj.Xana.get_item(ii)['trace'][:,obj.nq]
        tr[i] = np.vstack((np.mean(tmp,0), np.std(tmp,0))).T
    if trace:
        axtr = ax[1].twinx()
        niceplot(axtr)
        axtr.grid()
        axtr.set_ylabel('average intensity')
        axtr.set_yscale('log')

    ci = 0
    rates = np.zeros((len(obj.corrFunc), obj.nq.size, 2, 2))
    for j, (cfi, dcfi) in enumerate(corrFunc):
        ti = cfi[1:,0]
        if t_max > 0:
            t_ind = np.where(ti<=t_max)[0]
        elif t_max < 0:
            t_ind = np.arange(1,cfi.shape[0]+t_max)
        tf = np.logspace(np.log10(ti[0]), np.log10(ti[-1]),100)
        for i,qi in enumerate(obj.nq):
            if i == 0:
                cf_id = 'id {}'.format(obj.db_id[j])
            else:
                cf_id = None
            rates[j, i, 0, 1] = cfi[0,qi+1]
            rates[j, i, 1, :] = np.ma.average( cfi[t_ind+1,qi+1], weights=1/dcfi[t_ind+1,qi+1]**2, returned=1)
            rates[j, i, 1, 1] = np.sqrt(1/rates[j, i, 1, 1])
            ax[0].errorbar(ti, cfi[1:,qi+1], dcfi[1:,qi+1], label='{:.2e}nm-1'.format(obj.Xana.setup['qv'][qi]),
                        color=obj.colors[ci], marker=obj.markers[j%len(obj.markers)], markersize=2)
            ax[0].plot(tf, np.ones_like(tf)*rates[j, i, 1, 0], color=obj.colors[ci], )

            ax[1].errorbar(rates[j, i,0,1], rates[j, i,1,0], rates[j, i,1,1], color=obj.colors[ci],
                               label=cf_id, marker=obj.markers[j%len(obj.markers)], markersize=markersize)
            if trace:
                axtr.errorbar(rates[j, i,0,1], tr[j,qi,0], tr[j,qi,1], fmt='s', color=obj.colors[ci])
            ax[2].errorbar(tr[j,i,0], rates[j, i,1,0], xerr=tr[j,i,1], yerr=rates[j, i,1,1],
                                   color=obj.colors[ci], marker=obj.markers[j%len(obj.markers)],
                                   markersize=markersize)

            ci += 1
        staticContrast[j] = rates

    for axi in ax:
        niceplot(axi)
    ax[0].set_xscale('log')
    ax[1].legend(loc='best')
    ax[0].set_xlabel(r'delay time $\tau$ [s]')
    ax[0].set_ylabel(r'$g_2(\tau)$')
    ax[1].set_ylabel(r'contrast')
    ax[1].set_xlabel(r'q[nm-1]')

    ax[2].set_xscale('log')
    ax[2].set_xlabel('average intensity')
    ax[2].set_ylabel('contrast')

    plt.tight_layout()

    return staticContrast
    



import numpy as np

def gen_logind(limit, n):
    result = [1]
    if n>1:  # just a check to avoid ZeroDivisionError
        ratio = (float(limit)/result[-1]) ** (1.0/(n-len(result)))
    while len(result)<n:
        next_value = result[-1]*ratio
        if next_value - result[-1] >= 1:
            # safe zone. next_value will be a different integer
            result.append(next_value)
        else:
            # problem! same integer. we need to find next_value by artificially incrementing previous value
            result.append(result[-1]+1)
            # recalculate the ratio so that the remaining values will scale correctly
            ratio = (float(limit)/result[-1]) ** (1.0/(n-len(result)))
    # round, re-adjust to 0 indexing (i.e. minus 1) and return np.uint64 array
    return np.array(list(map(lambda x: round(x)-1, result)), dtype=np.uint64)



import numpy as np
import pyFAI
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator

def get_soq(Isaxs, mask, setup, Vsaxs=None, nbins=1000):
    ai = AzimuthalIntegrator(dist=setup['distance'],
                                       pixel1=setup['pix_size'][0]*1e-6,
                                       pixel2=setup['pix_size'][1]*1e-6)
    ai.setFit2D(setup['distance']*1000,
                      setup['ctr'][0],
                      setup['ctr'][1])
    ai.wavelength = setup['lambda']*1e-10
    if Vsaxs is None:
        q, ii, e = ai.integrate1d(Isaxs, nbins, mask=~(mask.astype(bool)),
                                  unit='q_nm^-1', error_model='poisson')
    else:
        q, ii, e = ai.integrate1d(Isaxs, nbins, mask=~(mask.astype(bool)),
                                  unit='q_nm^-1', variance=Vsaxs)
    return q, ii, e



import numpy as np
import lmfit
from .g2function import g2 as g2func
from copy import copy
import sys
import pandas as pd
import matplotlib
from matplotlib import pyplot as plt
import re
from scipy.special import gamma

class G2:

    def __init__(self, cf, nq, dcf=None):

        self.t = np.ma.masked_invalid(cf[1:,0])
        self.qv = cf[0,1:].copy()
        self.nq = np.asarray(nq)

        self.cf = np.ma.masked_invalid(cf[1:,1:]).T
        ind_sort = np.argsort(self.t)
        self.t = self.t[ind_sort]
        self.cf = self.cf[:,ind_sort]

        if isinstance(dcf, np.ndarray):
            self.dcf = np.ma.masked_invalid(dcf[1:,1:]).T
            self.dcf = self.dcf[:,ind_sort]
        else:
            self.dcf = None

        # properties
        self.nmodes = 1
        self.fitglobal = []
        self.fitqdep = None
        self.fix = None

        # attributes defined in private methods
        self._lmpars = None
        self._weights = None
        self._varnames = None

        self._fit_data = None
        self._fit_weights = None

        # fit results
        self.pars = None
        self.fit_result = []

        self._minimizer = None

        # plot handles
        self.plot_handles = None

    @property
    def nmodes(self):
        return self.__nmodes

    @nmodes.setter
    def nmodes(self, nmodes):
        if nmodes <= 0:
            print('Number of modes has been set to 1.')
            self.__nmodes = 1
        else:
            self.__nmodes = nmodes

    @property
    def fitglobal(self):
        return self.__fitglobal

    @fitglobal.setter
    def fitglobal(self, fitglobal):
        dofitglobal = bool(fitglobal)
        if dofitglobal:
            self.ndat = len(self.nq)
        else:
            self.ndat = 1
            fitglobal = []
        self.__fitglobal = fitglobal

    @staticmethod
    def _reduce_func(arr):
        return -0.5 * np.nansum(arr * arr)

    @staticmethod
    def _iter_cb(params, iter, resid, *args, **kws):
        if (iter) % 128 == 0:
            print('\rOptimizing...(%d)' % iter, end='', flush=1)
        else:
            pass


    def fit(self, mode='sig', nmodes=1, fitglobal=[], init={}, fix={}, lmfit_pars={}, fitqdep={}):
        '''
        Function that computes the fits using lmfit's minimizer
        '''

        self.fitglobal = fitglobal
        self.nmodes = nmodes
        self.fitqdep = fitqdep
        self.fix = fix

        # make the parameters
        if not lmfit_pars.get('is_weighted', True):
            init['__lnsigma'] = 1

        self._init_parameters(init, fix)

        # setting the weights of the data
        self._get_weights(mode)

        self._minimizer = lmfit.Minimizer(self._residuals, params=self._lmpars,
                                          reduce_fcn=self._reduce_func,
                                          iter_cb=self._iter_cb,
                                          nan_policy='omit')

        if bool(self.fitglobal):
            # set data to fit
            self._fit_data = self.cf[self.nq]
            self._fit_weights = self._weights[self.nq]

            # do the fit
            out = self._minimizer.minimize(**lmfit_pars)

            self._write_to_pars(out)
            self.fit_result.append((out, lmfit.fit_report(out)))

        else:
            for line, qi in enumerate(self.nq):
                # set data to fit
                self._fit_data = self.cf[qi]
                self._fit_weights = self._weights[qi]

                # do the fit
                out = self._minimizer.minimize( **lmfit_pars)

                self._write_to_pars(out, line=line)
                self.fit_result.append((out, lmfit.fit_report(out)))

        self.pars['q'] = self.qv[self.nq]
        self.pars = self.pars.apply(pd.to_numeric)

        return self.pars, self.fit_result

    def plot(self, doplot=False, marker='o', ax=None, xl=None, yl=None, colors=None, alpha=1.,
             markersize=6, markeredgecolor='w', data_label=None, confint=False, pars=None, **kwargs):

        sucfit = True if (self.pars is not None) else False

        if pars is not None :
            if len(pars):
                self.pars = pars
                self.nmodes = max([int(x[1]) if x.startswith('t') else 0 for x in pars.columns]) +1
                sucfit = True

        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(self.t)*0.5,np.max(self.t)*1.5)

        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]), 50)

        for ii, iq in enumerate(self.nq):
            pl = []
            if sucfit:
                ipars = np.abs(self.pars['q']-self.qv[iq]).idxmin()
                v = self.pars.iloc[ipars]
                g2f = 0
                for i in range(self.nmodes):
                    ve = f'{i}'
                    g2f += g2func(xf, t=v['t'+ve], b=v['b'+ve], g=v['g'+ve],
                                a=v['a']*1.*(not bool(i)))
                if confint:
                    g2fci = np.zeros((2,len(xf)))
                    for i in range(self.nmodes):
                        ve = f'{i}'
                        for j in range(2):
                            g2fci[j] += g2func(xf,
                                               t=v['t'+ve]*(1 + (-1)**j * .01 * confint.get('t'+ve, 0)),
                                               b=v['b'+ve]*(1 + (-1)**j * .01 * confint.get('b'+ve, 0)),
                                               g=v['g'+ve]*(1 + (-1)**j * .01 * confint.get('g'+ve, 0)),
                                               a=v['a']*1.*(not bool(i)))

                labstr_fit = None
                if 'legf' in doplot and sucfit:
                    pard = {'t':r'$t: {:.2e}\mathrm{{s}},\,$',
                            'g':r'$\gamma: {:.2g},\,$',
                            'b':r'$\mathrm{{b}}: {:.3g},\,$',
                            'a':r'$\mathrm{{a}}: {:.3g},\,$'}
                    labstr_fit = ''
                    for i in range(self.nmodes):
                        for vn in 'tgba':
                            if vn == 'a' and i > 0:
                                continue
                            elif vn == 'a' and i == 0:
                                vnn = 'a'
                            else:
                                vnn = vn + str(i)
                            if vnn in self.pars.columns:
                                labstr_fit += pard[vn].format(self.pars.loc[ii, vnn])
                            elif fix is not None and vnn in fix.keys():
                                labstr_fit += 'fix '+ pard[vn].format(fix[vnn])
                            else:
                                labstr_fit += pard[vn].format(0)

            labstr_data = None
            if 'legd' in doplot:
                if data_label is not None:
                    labstr_data = data_label
                else:
                    labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'legq' in doplot:
                labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'fit' in doplot and sucfit:
                if 'g1' in doplot:
                    g2f = np.sqrt((g2f-out.params['a'].value)/out.params['b0'].value)
                pl.append(ax.plot(xf, g2f, '-', label=labstr_fit, linewidth=1))
                if confint:
                    for j in range(2):
                        pl.append(ax.plot(xf, g2fci[j], ':', linewidth=1))

            if 'data' in doplot:
                if 'g1' in doplot:
                    cf = np.sqrt((cf-out.params['a'].value)/out.params['b0'].value)
                if self.dcf is not None:
                    pl.append(ax.errorbar(self.t.filled(np.nan), self.cf[iq].filled(np.nan), yerr=self.dcf[iq].filled(np.nan),
                                          linestyle='', marker=marker, label=labstr_data,
                                          alpha=alpha, markersize=markersize, mec=markeredgecolor, markeredgewidth=1))
                else:
                    pl.append(ax.plot(self.t, self.cf[iq], marker, label=labstr_data,
                                      alpha=alpha, markersize=markersize, mec=markeredgecolor, markeredgewidth=1))

            if colors is None:
                if 'data' in doplot:
                    color = pl[-1].get_color()
                else:
                    color = 'gray'
            else:
                color = colors[ii]

            for p in pl:
                if isinstance(p, list):
                    p[0].set_color(color)
                elif isinstance(p, matplotlib.container.ErrorbarContainer):
                    for child in p.get_children():
                        child.set_color(color)

            self.plot_handles = pl

            if 'g1' in doplot:
                ax.set_xscale('linear')
                ax.set_yscale('log')
            else:
                ax.set_xscale('log')
                ax.set_yscale('linear')

            if 'leg' in doplot:
                ax.legend()

            if 'report' in doplot and sucfit:
                print(lmfit.fit_report(out))

            ax.set_xlabel(r'$\tau\,(\mathrm{s})$')
            ax.set_ylabel(r'$g_2(\tau)$')

            ax.set_xlim(*xl)
            if yl is not None:
                ax.set_ylim(*yl)

        return None

    def _calc_model(self, v,):
        """Calculate multi mode g2 funktion
        """
        # v = pars.valuesdict()
        vn = self._varnames
        model = np.zeros((self.ndat,self.t.size), dtype=np.float64)

        for j in range(self.ndat):
            for i in range(self.nmodes):
                a = v[vn['a'][j][i]] if (i==0) else 0
                model[j] += g2func(self.t,
                                   t=v[vn['t'][j][i]],
                                   b=v[vn['b'][j][i]],
                                   g=v[vn['g'][j][i]],
                                   a=a)
        return model

    def _residuals(self, pars,):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        model = self._calc_model(v)

        resid = np.sqrt((self._fit_data - model)**2 * self._fit_weights**2)

        return np.squeeze(resid.flatten())


    def _init_parameters(self, init, fix):
        '''Initialize lmfit parameters dictionary
        '''

        self._initial_guess(init)

        # initialize parameters
        pars = lmfit.Parameters()
        for vn, vinit in init.items():
            if vn in self.fitqdep:
                pars.add(vn)
            else:
                pars.add(vn, value=vinit[0], min=vinit[1], max=vinit[2], vary=1)
        pars.add('a', value=init['a'][0], min=init['a'][1], max=init['a'][2], vary=1)
        pars.add('beta', value=init['beta'][0], min=init['beta'][1], max=init['beta'][2], vary=1)

        self._init_pars_dataframe(init)
        self._make_varnames()

        # setting contrast constraint
        beta_constraint = self._get_beta_constraint(ndat=-1)
        pars['b0'].set(expr=beta_constraint)

        # setting parameters fixed
        for vn in fix.keys():
            if vn in pars.keys():
                pars[vn].set(value=fix[vn], min=-np.inf, max=np.inf, vary=0)

        # modifying pars for global fitting if not globalfit then ndat=1
        lpars = list(pars.items())[::-1]
        for j in range(self.ndat-1):
            for pk, pv in sorted(lpars):
                if (pk not in self.fitglobal) or (pk in self.fitqdep):
                    pt = copy(pv)
                    pt.name += f'_{j}'
                    if pk == 'b0':
                        pb0 = pt
                        beta_constraint = self._get_beta_constraint(ndat=j)
                        pb0.set(expr=beta_constraint)
                        continue
                    # print(pars)
                    # print(pt)
                    pars.add(pt)

            # add the constraint for b0
            pars.add(pb0)

        re_par = re.compile('[tgb]\d{1}(?=_)')
        for p in self.fitqdep:
            for new_par, par_kw in self.fitqdep[p]['pars'].items():
                pars.add(new_par, **par_kw)
            for j in range(self.ndat):
                vn = p + f'_{j-1}'*bool(j)
                q = self.qv[self.nq[j]]
                expr = copy(self.fitqdep[p]['expr'])
                cpars = re_par.findall(expr)
                for c in cpars:
                    expr = expr.replace(f'{c}_', self._varnames[c[0]][j][int(c[1])])
                expr = expr.replace('q', f'{q:.5f}')
                pars[vn].set(expr=expr)


        if sum([x.startswith('t') for x in self.fitglobal]) == 0:
            for j in range(self.ndat):
                for i in range(1,self.nmodes):
                    vc = f't{i}' + f'_{j-1}' * bool(j)
                    vs = f't{i-1}' + f'_{j-1}'*bool(j)
                    pars.add(vs+'_dtmp', value=pars[vc].value - pars[vs].value, min=0)
                    pars[vc].set(expr=f'{vs} + {vs}'+'_dtmp')

        # print(pars)
        pars._asteval.symtable['gamma'] = gamma
        self._lmpars = pars

    def _get_weights(self, mode):
        '''
        mask data points to exclude them from the fit based on error bars and nan values
        and define weights for the fit
        '''
        dcf = self.dcf
        cf = self.cf.copy()
        if dcf is not None:
            excerr = (dcf.filled(0)<=0)
            wgt = dcf.copy()
            wgt = np.ma.masked_where(excerr, wgt)
            if mode == 'semilogx':
                wgt = np.log10(wgt)
            elif mode == 'semilogx2':
                wgt = 1/np.log10(wgt/cf)
            elif mode == 'semilogx3':
                wgt = 1/(cf * np.log10(wgt/cf))
            elif (mode == 'equal') or (mode == 'none') or (mode == None):
                wgt = np.ones_like(wgt)
            elif mode == 'logt':
                wgt = np.ones_like(cf)*np.log10(self.t)
            elif mode == 't':
                wgt = 1/self.t
            elif mode == 'sig**2':
                wgt = 1/wgt**2
            elif mode == 'sig':
                wgt = 1/wgt
            elif mode == 'logsig':
                wgt = 1/np.log10(wgt)
            elif mode == 'logsig+1':
                wgt = 1/np.log10(wgt+1)
            elif mode == 'data':
                wgt = 1/self.cf.copy()
            elif mode == 'test':
                wgt =  np.log10(self.t[0]/self.t+1) / np.log10(wgt+1)
            else:
                raise ValueError(f'Error mode {mode} not understood.')
        else:
            wgt = np.ma.masked_array(np.ones_like(self.cf))

        excdat = ~np.isfinite(self.cf) | wgt.mask
        self.cf = np.ma.masked_where(excdat, self.cf)

        self._weights = wgt

    def _init_pars_dataframe(self, init):
        """
        initialize pandas DataFrame for returning results
        """
        names = list(init.keys())
        cols = [names[i // 2] if (i + 1) % 2 else 'd' + names[i // 2]
                for i in range(len(names) * 2)]
        cols.insert(0, 'q')
        cols.extend(['chisqr', 'redchi', 'bic', 'aic'])
        self.pars = pd.DataFrame(columns=cols)

    def _initial_guess(self, init):
        """
        make initial guess for parameters if not provided by init
        """
        for i in range(self.nmodes):
            for s in 'tgb':
                vn = s+'{}'.format(i)
                if vn not in init.keys():
                    if s == 't':
                        t0 = np.logspace(np.log10(self.t.min()), np.log10(self.t.max()), self.nmodes+2)[i+1]
                        init[vn] = (t0, -np.inf, np.inf)
                    elif s == 'g':
                        init[vn] = (1, .2, 1.8)
                    elif s == 'b':
                        init[vn] = (.1, 0, 1)

        if 'a' not in init:
            init['a'] = (1, 0, 2)
        if 'beta' not in init:
            init['beta'] = (0.2, 0, 1)
        if bool(init.get('__lnsigma', False)):
            if not isinstance(init['__lnsigma'], tuple):
                init['__lnsigma'] = (np.log(0.1), None, None)

    def _get_beta_constraint(self, ndat=-1):
        dc = f'_{ndat}' * bool(ndat+1) # counter for fit_global
        dcb = '' if ('beta' in self.fitglobal) or ('beta' in self.fix) else dc
        dca = '' if ('a' in self.fitglobal) or ('a' in self.fix) else dc
        if self.nmodes > 1:
            beta_constraint = 'a' + dca + \
                              ' + beta' + dcb + \
                              ' - 1 - ' + '-'.join([f'b{x}' + dc for x in range(1,self.nmodes)])
        else:
            beta_constraint = 'beta' + dcb
        return beta_constraint

    def _make_varnames(self):
        """Make dict of varnames for easy handling in _calc_model
        """
        self._varnames = {'t': [],
                          'g': [],
                          'b': [],
                          'a': [],
                          'beta': [],
                          }

        if "__lnsigma" in self.pars.columns:
            self._varnames["__lnsigma"] = []

        for k in self._varnames.keys():
            cond = k in ['a', 'beta', '__lnsigma']
            for j in range(self.ndat):
                self._varnames[k].append([])
                jj = j - 1
                for i in range(self.nmodes):
                    ve = k + f'{i}' if not cond else k
                    if (ve not in self.fitglobal) or (ve in self.fitqdep):
                        ve += f'_{jj}' * bool(j)
                    self._varnames[k][j].append(ve)
                    if cond:
                        break

    def _write_to_pars(self, out, line=0):
        """ Save fit results in self.pars variable
        """
        for k in self._varnames.keys():
            cond = k in ['a', 'beta', '__lnsigma']
            gof = np.hstack([out.chisqr, out.redchi, out.bic, out.aic])
            for j in range(self.ndat):
                for i in range(self.nmodes):
                    df_name = k + f'{i}' if not cond else k
                    if (df_name not in self.fitglobal) or (df_name in self.fitqdep):
                        param_name = df_name + f'_{j-1}' * bool(j)
                    else:
                        param_name = df_name
                    value = out.params[param_name].value
                    try:
                        stderr = 1. * out.params[param_name].stderr
                    except TypeError:
                        stderr = np.nan
                    self.pars.loc[j+line, df_name] = value
                    self.pars.loc[j+line, 'd'+df_name] = stderr

                self.pars.iloc[j+line, -4:] = gof






#! /usr/bin/env python
import numpy as np
from numpy import log
from scipy.special import gamma


def PoissonGamma(x, M, p, ind_var='kb'):
    """Return Poisson-Gamma distribution as a function of k or kb. The other one is passed
    as a parameter. Depending on the argument ind_var. 
    """
    if ind_var == 'kb':
        kb = x
        k = p
    elif ind_var == 'k':
        k = x
        kb = p
    return gamma(1.*k+M)/(gamma(1.*M)*gamma(1.*k+1.))*1.*(1.*kb/(M*1.+kb))**k*(1.*M/(kb+M))**M

# def stirling(x):
#     return np.sqrt(2*np.pi/x)*(x/np.e)**x

# def PoissonGamma(x, **p):
#     k = p['k']
#     M = p['M']
#     return gamma(1.*k+M)/(gamma(1.*M)*gamma(1.*k+1.))*1.*(1.*x/(M*1.+x))**k*(1.*M/(x+M))**M

# def PoissonGamma_approx(x, **p):
#     k = p['k']
#     M = p['M']
#     return 1. -.5*log(2*np.pi*(k+M)/(M*(1+k))) + (k+M)*log((k+M)/(x+M)) + k*log(x) - (k+1)*log(1+k)    

# def PoissonGamma_indk(x, **p):
#     kb = p['kb']
#     M = p['M']
#     return gamma(1.*x+M)/(gamma(1.*M)*gamma(1.*x+1.))*1.*(1.*kb/(M*1.+kb))**x*(1.*M/(kb+M))**M

# def PoissonGamma_indk_approx(x, **p):
#     kb = p['kb']
#     M = p['M']
#     return 1. -.5*log(2*np.pi*(x+M)/(M*(1+x))) + (x+M)*log((x+M)/(kb+M)) + x*log(kb) - (x+1)*log(1+x)    



from time import time
import numpy as np
import pickle as pkl
from multiprocessing import Process, Queue
from .mp_corr3_err import mp_corr
from scipy.optimize import leastsq
from ..misc.progressbar import progress
import sys
from matplotlib import pyplot as plt


def rebin( a, newshape ):
    """Rebin an array to a new shape.
    """
    assert len(a.shape) == len(newshape)
    slices = [ slice(0,old, float(old)/new) for old,new in zip(a.shape,newshape) ]
    coordinates = np.mgrid[slices]
    indices = coordinates.astype('i')   #choose the biggest smaller integer index
    return a[tuple(indices)]

def errfunc(pa,xdata,ydata):
    """Fit function for fitting the variance of the two-time correlation function.
    """
    return (pa[0]+pa[1]*xdata-ydata)/np.sqrt(ydata*1e-8)

def trc(matr):
    """Calculate the two-time correlation function.
    """
    meanmatr = np.mean(matr, axis=1)
    meanmatr[meanmatr<=0] = 1.
    tmp, lenmatr = np.shape(matr)
    meanmatr.shape = 1, tmp
    trcm = np.dot(matr, matr.T) / lenmatr / np.dot(meanmatr.T, meanmatr)
    return trcm

def vartrc(ttc):
    """Calculate the variance of the two-time correlation function.
    """
    pc0 = [1.0,0.1]
    n,tmp = np.shape(ttc)
    vtmp = []
    for it in range(1,n-1):
        #ydata=diag(ttc,it)
        #xdata=arange(1,len(ydata)+1)
        #p1,success=leastsq(errfuncc,pc0,args=(xdata,ydata))
        #vtmp.append(var(ydata/(p1[0]+p1[1]*xdata)))              
        vtmp.append(np.var(np.diag(ttc,it)))         
    return vtmp

def recurf(ll):
    """Helper function used for calculating the two-time correlation function.
    """
    global l, y, v
    y[ll+1].append( (y[ll][0]+y[ll][1]) * 0.5)
    y[ll]=[]
    v[ll+1].append(vartrc(y[ll+1][-1]))
    if l[ll+1] == 1:
        recurf(ll+1)
    else:
        l[ll+1] += 1
    l[ll]=0  

def avr(saxs, ctr=-1, mask=None):
    """Old version of normalization function
    """
    dim1, dim2 = np.shape(saxs)

    if mask is None:
        mask = np.ones((dim1,dim2))
    saxs = saxs*mask
    
    if ctr == -1:
        return np.ones((dim1,dim2))*np.mean(saxs)
    cx,cy = ctr
    [X,Y] = np.mgrid[1-cy:dim1+1-cy,1-cx:dim2+1-cx]
    q = np.round(np.sqrt(X**2+Y**2)).astype(np.int64)

    q = q.ravel()
    mask = mask.flatten()
    saxs = saxs.flatten()

    qm = list(range(int(q.max()+1)))
    qr = list(range(len(qm)))
    for i in qm:
        qr[i] = []
    for i in range(len(q)):
        if (mask[i]):
            qr[q[i]].append(i)
    while [] in qr:
        qr.remove([])
    for i in qr:
        saxs[i] = np.mean(saxs[i])
    return saxs.reshape(dim1,dim2)

def avr_better( saxs, ctr, mask ):
    """Return an average saxs image for normalization of images.
    """
    cx, cy = ctr
    dim1, dim2 = np.shape(saxs)
    [X,Y] = np.mgrid[1-cy : dim1+1-cy, 1-cx : dim2+1-cx]
    q = np.float32(np.sqrt(X**2 + Y**2))
    n = np.int16(q + 0.5)
    q[mask==0] = 0
    n[mask==0] = 0
    max_n = n.max() + 1
    mean_saxs = np.zeros(max_n+1, np.float32)
    new_saxs = np.zeros_like(saxs, np.float32)
    radi = np.zeros((max_n,2), np.float32)

    for i in range(max_n):
        ind = np.where((n==i) & (mask==1))
        if ind[0].size:
            mean_saxs[i] = np.mean(saxs[ind])

    for i in range(dim1):
        for j in range(dim2):
            if q[i,j] > 0:
                par = int(q[i,j])
                f1 = q[i,j] - par
                if mean_saxs[par+1]>0 and mean_saxs[par]>0 :
                    new_saxs[i,j] = mean_saxs[par+1]*f1 + mean_saxs[par]*(1-f1) 
                if mean_saxs[par+1]>0 and mean_saxs[par]==0 :
                    new_saxs[i,j] = mean_saxs[par+1]
                if mean_saxs[par+1]==0 and mean_saxs[par]>0 :
                    new_saxs[i,j] = mean_saxs[par]
    return new_saxs

def calculate_twotime_correlation_function(ttdata, tt_max_images=5000):
    """Calculate two-time correlation function:
    The input ttdata has to be a numpy array of shape (nimages, npixels).
    """
    global l, v, y

    print("Start calculating TRC and Chi4...")

    nf, lind = ttdata.shape
    if nf > tt_max_images: # two-time correlation function is calculated for max tt_max_images
        ttchunk = nf // tt_max_images
        nfnew = ttchunk * tt_max_images
        print('Reducing two-time correlation data from {} to {} images by rebinning.'.format(nf,tt_max_images))
        ttdata = np.mean(ttdata[:nfnew].reshape(ttchunk,-1,lind, order='F'), 0)

    lind2 = lind // 16
    l = np.zeros(5)
    y=[]
    v=[]
    for i in range(5):
        y.append([])
        v.append([])

    ib = 0
    for i in range(16):
        ie = ib + lind2
        y[0].append(trc(ttdata[:,ib:ie]))
        v[0].append(vartrc(y[0][-1]))
        if l[0]==1:
            recurf(0)
        else:
            l[0] += 1
        ib += lind2

    vm = []
    for i in range(4,-1,-1):
        vm.append(np.mean(v[i],0))
    vm = np.array(vm)

    del ttdata
    del v

    ttcf = y[4][-1] # this is the two-time correlation function

    dia1 = np.mean(np.diag(ttcf,1))
    t = np.arange(np.shape(ttcf)[0])
    ttcf[t,t] = dia1

    N = np.array([1, 2, 4, 8, 16]) / float(lind)
    z = vm.T/N
    #p0=[0,1]
    #it=range(len(ttcf[1:,0]))
    #p1=zeros((len(ttcf[1:,0]),len(p0)+1))
    #p1[:,0]=(asfarray(it)+1.0)*dt 
    #xdata=ttcf[0,:]
    #for i in it:
    #    ydata=ttcf[i+1,:]
    #    p1[i,1:], success = leastsq(errfunc, p0, args=(xdata,ydata))

    return ttcf, z



#####################
#---MAIN FUNCTION---#
#####################
def pyxpcs( data, qroi, dt=1., qv=None, saxs=None, mask=1., ctr=(0,0), twotime_par=-1,
            qsec=(0,0), norm='symmetric_whole', nprocs=8, verbose=True, chn=16,
            tt_max_images=5000):
    """Calculate g2 correlation functions with a given dataset or chunks of a data set.
    """
    time0 = time()
    chn2 = int(chn/2)
    lqv = len(qroi)
    rlqv = range(lqv)

    if qv is None:
        qv = np.arange(lqv)

    if isinstance(data, np.ndarray):
        nf, *dim = np.shape(data)
        def get_chunk():
            return (0, data)
    elif isinstance(data, dict):
        nf = data['nimages']
        dim = data['dim']
        def get_chunk():
            return data['dataQ'].get()

    if verbose:
        print('Number of images is:', nf)
        print('shape of image section is:', dim)

    if isinstance(mask, np.ndarray):
        mask = mask[qsec[0]:qsec[0]+dim[0],qsec[1]:qsec[1]+dim[1]]
        if saxs is not None and saxs.shape!=mask.shape:
            saxs = saxs[qsec[0]:qsec[0]+dim[0],qsec[1]:qsec[1]+dim[1]]

   # normalize with average saxs image
    if saxs is not None:
        print('Start computing SAXS for normalization.')
        dim2, dim1 = np.shape(saxs)
        saxs_img = saxs.copy()
        ctr = (ctr[0]-qsec[1],ctr[1]-qsec[0])
        saxs_img = saxs_img * mask
        saxs_img = avr_better(saxs_img, ctr, mask)
        saxs_imgc = np.ones((dim2,dim1))
        saxs_img[saxs_img==0] = 1.
        for i in rlqv:
            q0 = qroi[i][0] - qsec[0]
            q1 = qroi[i][1] - qsec[1]
            saxs_imgc[q0,q1] = np.mean(saxs_img[q0,q1])/saxs_img[q0,q1]

        # saxs_imgc[np.where(np.isinf(saxs_imgc))] = 1.0
        print('Done')

        if verbose:
            print('Shape of saxs_img:',  np.shape(saxs_img))
            print('Sum of saxs_img:', np.sum(saxs_img))
    if verbose:
        print('Number of ROIs: ', lqv)

    lind = []
    total_pixels = 0
    for iq in rlqv:
        npixel = len(qroi[iq][0])
        lind.append(npixel)
        total_pixels += npixel

    nprocs = min(nprocs,lqv) # cannot use more processes than q-values
    tmp_pix = 0
    if nprocs >= lqv:
        q_sec = np.arange(lqv+1)
    else:
        q_sec = [lqv,]
        for iq in rlqv[::-1]:
            tmp_pix += lind[iq]
            if tmp_pix >= np.floor(total_pixels/nprocs):
                q_sec.append(iq)
                total_pixels -= tmp_pix
                nprocs -= 1
                tmp_pix =  0
            if iq == 0 or nprocs == 0:
                q_sec.append(0)
    q_sec = np.unique(q_sec)
    del tmp_pix
    nprocs = len(q_sec) - 1
    print('Using {} processes.'.format(nprocs))

    #----twotime----
    # writing data for two time correlatino function calculation to tttdata
    if twotime_par != -1:
        ttdata = np.empty((nf,lind[twotime_par]), dtype=np.float32)
    #-------------

    srch = int(np.ceil(np.log2(nf/chn))) + 1
    rcr =  int(chn + chn2*(srch-1))

    print('Number of registers is {} with {} total correlation points.'.format(srch, rcr))

    lag = np.zeros(rcr, dtype=np.float32)    # initialize lag time vector

    for ir in range(srch):
        if ir == 0:
            lag[:chn] = np.arange(1,chn+1)
        else:
            sl = slice(chn2*(ir+1),chn2*(ir+2))
            lag[sl] = 2**ir*np.arange(1+chn2,chn+1)

    rcrc = rcr - np.where(lag[sl]>nf)[0].size - 1
    lag *= dt # scale lag-vector with time step

    trace = np.empty((nf,lqv))
    tt_vec = np.linspace(0,nf,tt_max_images)*dt

    #----multiprocessing----
    # create lists of queues and processes
    qur = []
    qure = []
    pcorr = []
    for i in range(nprocs):
        qur.append(Queue(16))
        qure.append(Queue(1))
    for i in range(nprocs):
        q_beg = q_sec[i]
        q_end = q_sec[i+1]
        pcorr.append(Process(target=mp_corr, args=(i, nf-1, chn, srch, rcr,
                lind[q_beg:q_end], q_end-q_beg, qur[i], qure[i])))

    # start processes
    for i in range(nprocs):
        pcorr[i].start()
    #-----------------------

    # start processing data
    tcalc_cum = 0
    t0 = 0
    last_chunk = -1
    lin_mask = np.where(mask)
    while t0 < nf - 1:
        progress(t0,nf)

        c_idx, chunk = get_chunk()
        chunk_size = chunk.shape[0]
        idx = slice(t0,t0+chunk_size)
        # matr[matr<0] = 0

        chunk_diff = c_idx - last_chunk
        if chunk_diff != 1:
            raise IOError('Chunks have been read in wrong order: chunk index difference is % and not 1.' %
                             chunk_diff)
        last_chunk = c_idx

        if saxs is not None:
            chunk = chunk * saxs_imgc # normalize with mean saxs image

        # save data for two time correlation
        if twotime_par != -1:
            ttdata[idx,:] = chunk[:,qroi[twotime_par][0]-qsec[0],qroi[twotime_par][1]-qsec[1]]

        for jj,(i,j) in enumerate(zip(q_sec[:-1], q_sec[1:])):
            tmp_put = []
            for qi in range(i,j):
                q0 = qroi[qi][0] - qsec[0]
                q1 = qroi[qi][1] - qsec[1]
                trace[idx,qi] = chunk[:,q0,q1].mean(-1)
                if norm == 'symmetric':
                    normfactor = trace[idx,qi].copy()
                    normfactor[normfactor==0] = 1.
                    tmp_put.append(chunk[:,q0,q1]/normfactor[:,None])
                elif norm == 'symmetric_whole':
                    tmp_put.append(chunk[:,q0,q1]/np.mean(chunk[:,lin_mask[0],lin_mask[1]], axis=1)[:,None])
                elif norm == 'none':
                    tmp_put.append(chunk[:,q0,q1])
                elif norm == 'corrcoef':
                    tmp_mat = chunk[:,q0,q1]/trace[idx,qi,None]
                    tmp_put.append((tmp_mat-tmp_mat.mean(-1)[:,None])
                                   /np.sqrt(np.var(tmp_mat,-1))[:,None])
            qur[jj].put(tmp_put)

        t0 += chunk_size

    progress(1,1)

    # read data from output queue
    from_proc = []
    for i in range(nprocs):
        from_proc.append(qure[i].get())
        pcorr[i].join()
        qure[i].close()
        qure[i].join_thread()

    # get correlation functions and normalization from processes
    corf = from_proc[0][0]
    dcorf = from_proc[0][1]
    nk = from_proc[0][2]
    sr = from_proc[0][3]
    sl = from_proc[0][4]
    tcalc_cum = from_proc[0][5]
    for i in range(1,nprocs):
        corf = np.concatenate((corf,from_proc[i][0]), axis=1)
        dcorf = np.concatenate((dcorf,from_proc[i][1]), axis=1)
        sr = np.concatenate((sr,from_proc[i][3]), axis=1)
        sl = np.concatenate((sl,from_proc[i][4]), axis=1)
        tcalc_cum = max(tcalc_cum,from_proc[i][5])

    if norm in ['symmetric', 'sym_!trace', 'symmetric_whole', 'none']:
        tmp = nk[:rcrc,None]**2/(sr[:rcrc]*sl[:rcrc])
        corf = corf[:rcrc] * tmp
        dcorf = np.abs(dcorf[:rcrc] * tmp**2 / (nk[:rcrc,None]**2))
    elif norm == 'corrcoef':
        corf = corf[:rcrc]
        dcorf = np.abs(dcorf[:rcrc])

    # initialize correlation array 'cc'
    cc = np.zeros((rcrc+1,lqv+1), dtype=np.float32)
    dcc = cc.copy()
    cc[1:,0] = lag[:rcrc]
    cc[1:,1:] = corf
    dcc[1:,0] = lag[:rcrc]
    dcc[1:,1:] = np.sqrt(dcorf)
    cc[0,1:] = qv
    dcc[0,1:] = qv

    if verbose:
        print("\rFinished calculating correlation functions.")

    del corf, dcorf, rcr, srch, lag, from_proc


    #----twotime and chi4----
    tcalc_cumtrc0 = time()
    if twotime_par != -1:
        ttcf, chi4 = calculate_twotime_correlation_function(ttdata, tt_max_images)
    else:
        ttcf = chi4 = 0

    tcalc_cumtrc = time()-tcalc_cumtrc0
    if verbose:
        print('Elapsed time: {:.2f} min'.format((time()-time0)/60.))
        print('Elapsed time for correlate: {:.2f} min'.format(tcalc_cum/60.))
        print('Elapsed time for TRC and Chi4: {:.2f} min'.format(tcalc_cumtrc/60.))

    corfd = {'corf':cc,
             'dcorf':dcc,
             'trace':trace,
             'qv':qv,
             'qroi':qroi,
             'Isaxs':saxs,
             'mask':mask,
             'twotime_corf':{twotime_par:ttcf},
             'twotime_par':twotime_par,
             'twotime_xy':tt_vec,
             'chi4':chi4
    }

    return corfd



from time import time
import warnings
import numpy as np
import sys
from scipy.ndimage import gaussian_filter
from ..misc.progressbar import progress
from .xpcsmethods import cftomt, mat2evt

try:
    from .fecorrt3m import fecorrt3m
except ImportError:
    warnings.warn("Could not load fortran module fecorrt3m. \
    Probably not built.")

#---MAIN FUNCTION---
def eventcorrelator(data, qroi, qv=None, dt=1., method='matrix',
                    twotime_par=-1, **kwargs):
    '''
    Event correlator
    '''
    time0 = time()
    lqv = len(qroi)
    rlqv = range(lqv)
    twotime_par = np.array(twotime_par)

    if qv is None:
        qv = np.arange(lqv)

    for roii in rlqv:
        print('\nAnalyzing ROI: {}'.format(roii), flush=1)
        if method == 'matrix':
            roi = data[:,qroi[roii][0],qroi[roii][1]]
            ntimes, npix = np.shape(roi)
            pix, t, s = mat2evt(roi)
        elif method == 'events':
            npix = qroi[roii][0].size
            pix, t, s = data[roii]
            ntimes = s.size
            # # old code
            # ntimes = len(data)
            # npix = qroi[roii][0].size
            # pix, t, s = [], [], []
            # for i, dati in enumerate(data):
            #     lpix = len(dati[roii])
            #     pix.append(dati[roii])
            #     t.append(np.zeros(lpix)+i)
            #     s.append(lpix)
            # pix = np.concatenate(pix)
            # t = np.concatenate(t)
            # s = np.asarray(s)

        if roii == 0:
            ttcf = {}
            cfmt = []
            trace = []
            cc = np.zeros((ntimes,lqv+1), np.float32)
            tt = np.arange(1,ntimes+1)*dt
            cc[0,1:] = qv
            z = cc.copy()

        indpi = np.argsort(pix)
        t = t[indpi]
        pix = pix[indpi]

        lpi = len(pix)
        cor = np.zeros((ntimes, ntimes))
        print('starting fortran routine', flush=1)
        cor = fecorrt3m(pix, t, cor, lpi, ntimes)
        lens = len(s)
        s = s.astype(dtype=np.float32)
        cor = np.array(cor, dtype=np.float32)
        s.shape = (lens, 1)
        norm = np.dot(s, np.flipud(s.T)) / ntimes
        cor = cor / norm * npix / ntimes
        tmp = np.mean(np.diag(cor, k=1))
        for i in range(ntimes - 1):
            cor[i,i] = tmp

        x = np.ones((ntimes-1,3))
        x[:,0] = np.arange(1,ntimes)
        for i in range(1,ntimes-1):
            dia = np.diag(cor,k=i)
            ind = np.where(np.isfinite(dia))
            x[i-1,1] = np.mean(dia[ind])
            x[i-1,2] = np.std(dia[ind])
        x[:,2] *= np.sqrt(1.0/(ntimes-1))
        x[:,0] *= dt
        cc[1:,roii+1] = x[:,1]
        z[1:,roii+1] = x[:,2]**2
        if roii == 0:
            cc[1:,0] = x[:,0]
            z[1:,0] = x[:,0]
        del x
        cfmt.append(cftomt(cc[1:,[0,roii+1]],err2=z[1:,roii+1]))
        trace.append(s)
        
        if roii in twotime_par:
            ttcf[roii] = cor.copy()
        del cor

    shp = cfmt[0].shape[0]
    corf = np.empty((shp+1,len(qv)+1))
    corf[1:,0] = cfmt[0][:,0]
    corf[0,1:] = qv
    dcorf = corf.copy()
    for i in rlqv:
        corf[1:,i+1] = cfmt[i][:,1]
        dcorf[1:,i+1] = cfmt[i][:,2]

    trace = np.squeeze(np.array(trace)).T        

    corfd = {'corf':corf,
             'dcorf':dcorf,
             'corf_full':cc,
             'dcorf_full':z,
             'trace':trace,
             'qv':qv,
             'qroi':qroi,
             'twotime_corf':ttcf,
             'twotime_xy':tt
    }
    return corfd



import numpy as np
import time
import copy
import multiprocessing as mp
from multiprocessing.managers import SyncManager
from queue import PriorityQueue
from .XpcsAna.Xpcs import Xpcs
from .XsvsAna.Xsvs import Xsvs
from .SaxsAna.Saxs import Saxs
from .ProcData.Xdata import Xdata
from .Decorators import Decorators
from .misc.xsave import save_result


class MyManager(SyncManager):
    pass


def Manager():
    m = MyManager()
    m.start()
    return m


class Analysis(Xdata):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

    @Decorators.input2list
    def analyze(self, series_id, method, first=0, last=np.inf, handle_existing='next',
                nread_procs=1, chunk_size=200, verbose=True, dark=None,
                dtype=np.float32, filename='', read_kwargs={}, **kwargs):

        if not self.setup.wavelength:
            raise ValueError('Setup is not defined properly. Cannot perform analysis.')

        for sid in series_id:
            if verbose:
                print('\n\n#### Starting %s Analysis ####\nSeries: %d in folder %s\n' %
                      (method, sid, self.datdir))
                print('Using {} processes to read data.'.format(nread_procs))

            # copy the metadata
            self._meta_save = copy.deepcopy(self.meta)
            rois = copy.deepcopy(self.setup.qroi)

            # if dark is not None:
            #     if type(dark) == int:
            #         print('Loading DB entry {} as dark.'.format(dark))
            #         dark = self.xana.get_item(dark)['Isaxs']

            nf = self.meta.loc[sid, 'nframes']
            first_proc = first % nf + self.meta.loc[sid, 'first']
            last_proc = min([self.meta.loc[sid, 'nframes'], last])
            last_proc = (last_proc - 1) % nf + self.meta.loc[sid, 'first']
            self._meta_save.loc[sid, ['first', 'last', 'nframes']] = (first_proc,
                                                                     last_proc,
                                                                     last_proc - first_proc + 1)

            # update meta database
            # self.meta.loc[sid, 'first'] = first
            # self.meta.loc[sid, 'last'] = last

            # dict with options and variables passed to the data reader
            read_opt = {'first': first_proc,
                        'last': last_proc,
                        'dark': dark,
                        'verbose': False,
                        'dtype': dtype,
                        'qsec': self.setup.qsec,
                        'output': '2dsection',
                        'nprocs': nread_procs,
                        'chunk_size':chunk_size
                        }
            saxs_dict = read_opt.copy()
            read_opt.update(read_kwargs)

            proc_dat = {'nimages': self._meta_save.loc[sid, 'nframes'],
                        'dim': self.setup.qsec_dim
                        }

            # old chunks
            # chunks = [np.arange(first + i*chunk_size, min(first + (i + 1)*chunk_size, last))
            #           for i in range(np.ceil((last - first) / chunk_size).astype(np.int32))]


            # new chunks
            ind_arange = np.arange(first_proc,last_proc+1)
            bins = np.arange(0, nf, chunk_size)
            digitized = np.digitize(ind_arange, bins)
            chunks = [ind_arange[np.where(digitized==i)] for i in np.unique(digitized)]

            if method in ['xpcs', 'xsvs']:

                # Register a shared PriorityQueue
                MyManager.register("PriorityQueue", PriorityQueue)
                m = Manager()
                dataQ = m.PriorityQueue(nread_procs)
                indxQ = m.PriorityQueue()
                #dataQ = mp.Queue(nread_procs)
                #indxQ = mp.Queue()'symmetric_whole'

                # add queues to read and process dictionaries
                read_opt['dataQ'] = dataQ
                read_opt['indxQ'] = indxQ
                read_opt['method'] = 'queue_chunk'
                proc_dat['dataQ'] = dataQ

                for i, chunk in enumerate(chunks):
                    indxQ.put((i, chunk))

                # h5 files can only be opened by one process at a time and, therefore,
                # the processes have to acquire a lock for reading data
                lock = 0
                if 'h5' in self.fmtstr:
                    lock = mp.Lock()
                    read_opt['lock'] = lock

                procs = []
                for ip in range(nread_procs):
                    procs.append(mp.Process(target=self.get_series,
                                            args=(sid,), kwargs=read_opt))
                    procs[ip].start()
                    time.sleep(2)

            if method == 'xpcs':
                saxs = kwargs.pop('saxs', 'compute')
                Isaxs = self.get_xpcs_args(sid, saxs, saxs_dict)
                dt = self.get_delay_time(sid)
                savd = Xpcs.pyxpcs(proc_dat, rois, dt=dt, qv=self.setup.qv,
                                   saxs=Isaxs, mask=self.setup.mask, ctr=self.setup.center,
                                   qsec=self.setup.qsec[0], **kwargs)

            elif method == 'xpcs_evt':
                dt = self.get_delay_time(sid)
                evt_dict = dict(method='events',
                                verbose=True,
                                qroi=rois,
                                dtype=np.uint32,
                )
                read_opt.update(evt_dict)
                evt = self.get_series(sid, **read_opt)
                savd = Xpcs.eventcorrelator(evt[1:], rois, self.setup.qv,
                                            dt, method='events', **kwargs)

            elif method == 'xsvs':

                t_e = self.get_xsvs_args(sid,)
                savd = Xsvs.pyxsvs(proc_dat, rois, t_e=t_e,
                                   qv=self.setup.qv, qsec=self.setup.qsec[0],
                                   **kwargs)

            elif method == 'saxs':

                read_opt['output'] = '2d'
                proc_dat = {'get_series': self.get_series,
                            'sid': sid,
                            'setup': self.setup,
                            'mask': self.setup.mask}
                savd = Saxs.pysaxs(proc_dat, **read_opt, **kwargs)

            else:
                raise ValueError('Analysis type %s not understood.' % method)

            if method in ['xpcs', 'xsvs']:
                # stopping processes
                for ip in range(nread_procs):
                    procs[ip].join()

                # closing queues
                # dataQ.close()
                # dataQ.join_thread()
                # indxQ.close()
                # indxQ.join_thread()

            f = self.datdir.split('/')[-2] + '_s' + \
                str(self.meta.loc[sid, 'series']) + filename
            savfile = save_result(
                savd, method, self.savdir, f, handle_existing)

            self.add_db_entry(sid, savfile, method)

    def get_xpcs_args(self, sid, saxs, read_opt):
        ''' Get Saxs and delay time for XPCS analysis.
        '''
        if saxs == 'compute':
            print('Calculating average SAXS image.')
            Isaxs = self.get_series(sid, method='average', **read_opt)[0]
        elif isinstance(saxs, int):
            if saxs == -1:
                saxs = self.db.shape[0] - 1
            print('Loading average SAXS from database entry {}'.format(saxs))
            Isaxs = self.get_item(saxs)['Isaxs']
        else:
            Isaxs = saxs
        return Isaxs

    def get_delay_time(self, sid):
        dt = 0
        for attr in ['t_delay', 't_exposure', 't_readout', 't_latency', 'rate',
                     'pulseLength']:
            if attr in self.meta.columns:
                item = self.meta.loc[sid, attr]
                if attr == 'rate':
                    dt += 1/item
                elif attr == 'pulseLength':
                    dt += item * 1e-15
                else:
                    dt += item
                    if attr == 't_delay':
                        break
        return dt

    def get_xsvs_args(self, sid):
        ''' Get exposure time for XSVS analysis
        '''
        t_e = 0
        for attr in ['t_exposure', 'pulseLength']:
            if attr in self.meta.columns:
                item = self.meta.loc[sid, attr]
                if attr == 'pulseLength':
                    t_e += item * 1e15
                else:
                    t_e += item

        return t_e

    def defineqrois(self, input_, **kwargs):
        if type(input_) == int:
            Isaxs = self.get_item(input_)['Isaxs']
        elif type(input_) == np.ndarray:
            Isaxs = input_
        if Isaxs.ndim == 3:
            Isaxs = self.arrange_tiles(Isaxs)
        Saxs.defineqrois(self.setup, Isaxs, **kwargs)

    @staticmethod
    def find_center(*args, **kwargs):
        return Saxs.find_center(*args, **kwargs)



import os
import numpy as np
import multiprocessing as mp
from readDet.openFile_mp import getEDFimagedata
import time

def rebin(arr, new_shape):
    """Rebin 2D array arr to shape new_shape by averaging."""
    shape = (new_shape[0], arr.shape[0] // new_shape[0],
             new_shape[1], arr.shape[1] // new_shape[1])
    return arr.reshape(shape).mean(-1).mean(1)


def make_spacer(nf):
    chn = 16
    chn2 = int(chn/2)
    srch = int(np.ceil(np.log2(nf/chn))) + 1
    rcr =  int(chn + chn2*(srch-1)) 

    lag = np.zeros(rcr, dtype=np.float32)    # initialize lag time vector

    for ir in range(srch):
        if ir == 0:
            lag[:chn] = np.arange(1,chn+1)
        else:
            sl = slice(chn2*(ir+1),chn2*(ir+2))
            lag[sl] = 2**ir*np.arange(1+chn2,chn+1)

    rcrc = rcr - np.where(lag[sl]>nf)[0].size - 1
    return lag[:rcrc].astype(np.uint16) 

def xsvs_roi(imgr):
    global nb, nb2
    kb = np.round(np.mean(imgr))
    ne = np.max([0,np.floor(np.log2(np.max([kb,1])/nb2))])
    if (kb <= 0) or (ne == 0):
        kbr = np.arange(nb+1)-0.5
        bc = kbr[:-1]+0.5
    elif (kb > 0) and (ne > 0):
        kbr = np.arange(0,(2**ne+1)*(nb+1),2**ne+1)-0.5
        bc = kbr[:-1]+2**(ne-1)+0.5
        
    nt, gp = np.shape(imgr)
    sr = np.zeros((nt+1,nb+1))
    dsr = sr.copy()
    sr[0,1:] = bc
    dsr[0,1:] = bc
    
    for j in range(nt):
        imgi = imgr[j,:]
        tmp = np.append(np.sum(imgi), np.histogram(imgi, kbr)[0])
        sr[j+1,:] = tmp / gp
        dsr[j+1,:] = np.sqrt(tmp) / gp
        
    return sr, dsr

def calc_xsvs_pproc(img, qroi, pool):
    global nb
    last = np.shape(img)[0]
    roi = np.zeros((len(qroi), last+1, nb+1), np.float32)
    droi = np.zeros_like(roi)
    
    workin = []
    for qi in qroi:
        workin.append(img[:,qi[0],qi[1]])
    del img
    
    for i, res in enumerate(pool.map(xsvs_roi, workin)):
        roi[i] = res[0]
        droi[i] = res[1]
    return roi, droi

def sum_data(data, spacer, nt, method):
    # Sum images to calculate Pairs or Joerg function
    sumdata = np.zeros((nt,*data.shape[:-1]), dtype=np.float32)
    for image_num in range(nt):
        imi = image_num
        if method == 2:
            imf = image_num + spacer
            S = np.sum(data[:,:,imi:imf], axis=-1)
        elif method == 1:
            imf = image_num + spacer - 1
            if imi == imf:
                S = data[:,:,imi]
            else:
                S = data[:,:,imi] + data[:,:,imf]
        sumdata[:,:,image_num] = S
    return sumdata

def sum_events(events, spacer, nt, method):
    """Sum events to calculate Pairs or Joerg function"""
    nq = len(events[0])
    sumevents = [[] for i in range(nt)] 
    for image_num in range(nt):
        imi = image_num
        if method == 2:
            imf = image_num + spacer
            S = [np.hstack(events[imi:imf][iq]) for iq in range(nq)] 
        elif method == 1:
            imf = image_num + spacer - 1
            if imi == imf:
                S = events[imi]
            else:
                S = [np.append(events[imi][iq],events[imf][iq]) for iq in range(nq)] 
        sumevents[image_num] = S
    return sumevents

def evt2prob( events, qroi, nbins ):
    ltimes = len(events)
    nq = len(events[0])
    prob = np.empty((nq,ltimes+1,nbins+1), np.float32)
    kvec = np.append(0,np.arange(nbins))
    for it in range(ltimes):
        for iq in range(nq):
            counts = np.unique(events[it][iq], return_counts=1)[1]
            tmp = np.histogram(counts, nbins-1, range=(1,nbins-1))[0]
            npix = qroi[iq][0].size
            prob[iq,0,:] = kvec
            prob[iq,it+1,:] = np.hstack((counts.sum(),npix-counts.size,tmp))/npix
    return prob, prob*0.


#--MAIN FUNCTION---
def pyxsvs( datin, qroi, last=None, method='xsvs', filen=0, data_format='edfid10',
                data_form='events', max_spacer=1, nspacer=1,
                dt=(1.,1.), nbins=15, nproc=4):
    
    time_i = time.time()
    
    provided_methods = {'xsvs':0, 'pairs':1, 'joerg':2}
    method = provided_methods[method]

    readmethods = {'edfid10':getEDFimagedata,}
    getDataFunc = readmethods[data_format]

    if nproc:
        pool = mp.Pool(processes=nproc)

    #initialize output array
    xsvs = []
    dxsvs = []
    
    if method == 0:
        lags = tv = None
        nfiles = len(datin)
        for n, dati in enumerate(datin):
            if data_form == 'events':
                roi, droi = evt2prob(dati, qroi, nbins)
            else:
                img = getDataFunc(dati, output=0)
                roi, droi = calc_xsvs_pproc(img, qroi, pool)
            xsvs.append(roi)
            dxsvs.append(droi)
                
    elif method:
        datin = datin[filen]
        if data_form == 'events':
            levt = len(datin)
            lags = make_spacer(levt)
            print('Using lags: {}'.format(lags))
            for inds,spacer in enumerate(lags):
                nt = levt - spacer + 1
                roi = np.zeros((len(qroi), nt+1, nbins+1), np.float32)
                droi = np.zeros_like(roi)

                Sevt = sum_events(datin, spacer, nt, method)
                roi, droi = evt2prob(Sevt, qroi, nbins)

                xsvs.append(roi)
                dxsvs.append(droi)
        else:
            img = getDataFunc(datin, last=(last,), nprocs=4, dtype=np.int32)
            print('Data loaded.', flush=1)
            for inds,spacer in enumerate(lags):
                nt = img.shape[0] - spacer + 1
                roi = np.zeros((len(qroi), nt+1, nbins+1), np.float32)
                droi = np.zeros_like(roi)

                Sdata = sum_data(img, spacer, nt, method)
                roi, droi = calc_xsvs_pproc(Sdata, qroi, pool)
                xsvs.append(roi)
                dxsvs.append(droi)
                    
        tv = lags * dt[0] + (lags-1)*dt[1]

    if nproc:
        pool.close()
        pool.join()

    time_f = time.time() - time_i
    print('Elapsed time: {}min'.format(round(time_f/60,2)))
    return {'xsvs':xsvs, 'dxsvs':dxsvs, 'time':tv, 'spacer':lags}
        



import numpy as np
import lmfit
from copy import copy
import sys
import pandas as pd
import matplotlib
from matplotlib import pyplot as plt
import re
from scipy.interpolate import splev, splrep
from numpy.fft import fft
from scipy.special import gamma
from scipy.signal import savgol_filter

def rescale(y, mn, mx, rng=(0,1)):
    p = (rng[1]-rng[0])/(mx-mn)
    return p * (y - mn) + rng[0], p

def rescale_baseline(y, index=None, rng=[0,1], npoints=10):
    rng[1] = np.max(y)
    mx = rng[1]
    if index is None:
        index = -1
    mn = np.nanmean(y[-npoints:,index])
    p = (rng[1]-rng[0])/(mx-mn)
    return p * (y - mn) + rng[0], p


def read_rheo_data(filename):
    return pd.read_table(filename, sep=';', na_values=' ')

def get_g1(g2, index=0):
    g1 = g2.corrFuncRescaled[index][0].copy()
    dg1 = g2.corrFuncRescaled[index][1].copy()
    ind = (slice(1,g1.shape[0]),g2.nq+1)
    g1[ind] -= 1
    g1[g1<0] = np.nan
    norm = np.asarray(g2.pars[index].loc[0, 'beta'])
    g1[ind] /= norm
    dg1[ind] /= norm
    g1[ind] = np.sqrt(g1[ind])
    dg1[ind] = .5 / g1[ind] * dg1[ind]
    return g1, dg1

def get_msd(g1, dg1=None):
    msd = g1.copy()
    msd[1:,1:] = -np.log(msd[1:,1:])*6 / msd[0,1:]**2
    if dg1 is None:
        return msd
    else:
        dmsd = dg1.copy()
        dmsd[1:,1:] = np.abs(1/g1[1:,1:]*6/g1[0,1:]**2*dg1[1:,1:])
        return msd, dmsd

def bspline_msd(x, y, plot=False, **kwargs):
    """Interpolate y=f(x) with Bsplines of degree <degree>
       and smoothness factor <s>
    """
    ind = np.where(~np.isnan(y)&(y>0))
    x = np.log(x[ind])
    y = np.log(y[ind])
    if 'w' in kwargs:
        kwargs['w'] = np.log(kwargs['w'][ind])

    spl = splrep(x, y, **kwargs)
    xx = np.linspace(x.min(), x.max(), 200)

    y_i = splev(xx, spl)
    if plot:
        fig, ax = plt.subplots()
        ax.plot(x, y, '-', label='data')
        ax.plot(xx, y_i, 'b-', lw=4, alpha=0.7, label='BSpline')
        ax.grid(True)
        ax.legend(loc='best')
        ax.legend(loc='lower left', ncol=2)
        plt.show()

    x = np.ma.masked_invalid(np.exp(xx))
    y = np.ma.masked_invalid(np.exp(y_i))

    return x, y

def savgol_msd(x, y, plot=False, **kwargs):
    """Interpolate y=f(x) with Bsplines of degree <degree>
       and smoothness factor <s>
    """
    ind = np.where(~np.isnan(y)&(y>0))
    x = np.log(x[ind])
    y = np.log(y[ind])
    if 'w' in kwargs:
        kwargs['w'] = kwargs['w'][ind]

    xx = np.linspace(x.min(), x.max(), 200)

    y_i = savgol_filter(y, 7, 4)

    return np.exp(x), np.exp(y_i)

def first_derivative(x,y):
    dy = np.zeros(y.shape)
    dy[0:-1] = np.diff(y)/np.diff(x)
    dy[-1] = (y[-1] - y[-2])/(x[-1] - x[-2])
    return dy

def calc_gomega(x,y,timestep=1):
    dy = first_derivative(np.log(x),np.log(y))
#     dy[np.where(dy<=0)[0][0]:] = dy[np.where(dy<=0)[0][0]-1]
#     dy = np.log(dy)
    x = np.log(x)
    dyfft = fft(dy)
    gom = 1/dyfft
    freq = np.fft.fftfreq(x.shape[-1], np.mean(np.diff(x)))
    return freq, gom

def calc_viscel_spec(x,y):
    dy = first_derivative(np.log(x),np.log(y))
    Gs = 1./(y*gamma(1+dy))
    return 1/x, Gs

def calc_complex_shear_modulus(x,y):
    alpha = first_derivative(np.log(x),np.log(y))
    G = 1./(y*gamma(1+alpha))
    Gp = G*np.cos(np.pi*alpha/2)
    Gpp = G*np.sin(np.pi*alpha/2)
    return 2*np.pi/(x), G, Gp, Gpp

class MSD:

    def __init__(self, msd):

        self.t = np.ma.masked_invalid(msd[0])
        ind_sort = np.argsort(self.t)
        self.t = self.t[ind_sort]

        self.msd = np.ma.masked_invalid(msd[1,ind_sort])
        self.dmsd = np.ma.masked_invalid(msd[2,ind_sort])

        # properties
        self.fix = None

        # attributes defined in private methods
        self._lmpars = None
        self._weights = None
        self._varnames = None

        self._fit_data = None
        self._fit_weights = None
        self._fit_logarithmic = False

        # fit results
        self.pars = None
        self.fit_result = []

        self._minimizer = None

    @staticmethod
    def _reduce_func(arr):
        return -0.5 * np.nansum(arr * arr)

    @staticmethod
    def _iter_cb(params, iter, resid, *args, **kws):
        if (iter) % 4 == 0:
            print('\rOptimizing...(%d)' % iter, end='', flush=1)
        else:
            pass


    def fit(self, mode='sig', init={}, fix={}, lmfit_pars={}, fitqdep={}):
        '''
        Function that computes the fits using lmfit's minimizer
        '''

        self.fix = fix

        # make the parameters
        if not lmfit_pars.get('is_weighted', True):
            init['__lnsigma'] = 1

        self._init_parameters(init, fix)

        # setting the weights of the data
        self._fit_data = self.msd
        self._get_weights(mode)
        self._fit_weights = self._weights


        self._minimizer = lmfit.Minimizer(self._residuals, params=self._lmpars,
                                          reduce_fcn=self._reduce_func,
                                          iter_cb=self._iter_cb,
                                          nan_policy='omit')

        # do the fit
        out = self._minimizer.minimize(**lmfit_pars)

        self._write_to_pars(out,)
        self.fit_result.append((out, lmfit.fit_report(out)))

        # convert to numeric data types
        self.pars = self.pars.apply(pd.to_numeric)

        self._fit_logarithmic = False

        return self.pars, self.fit_result


    def plot(self, doplot=False, marker='o', ax=None, xl=None, yl=None, colors=None, alpha=1.,
             markersize=3., data_label=None, confint=False, pars=None, **kwargs):

        sumidit = True if (self.pars is not None) else False

        if pars is not None :
            if len(pars):
                self.pars = pars
                sumidit = True

        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(self.t)*0.5,np.max(self.t)*1.5)

        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]), 50)

        for ii, iq in enumerate([0]):
            pl = []
            if sumidit:
                msdf = self._calc_model(self.fit_result[ii][0].params, x=xf)

                labstr_fit = None
                if 'legf' in doplot and sumidit:
                    pard = {'t':r'$t: {:.2e}\mathrm{{s}},\,$',
                            'g':r'$\gamma: {:.2g},\,$',
                            'b':r'$\mathrm{{b}}: {:.3g},\,$',
                            'a':r'$\mathrm{{a}}: {:.3g},\,$'}
                    labstr_fit = ''
                    for i in range(self.nmodes):
                        for vn in 'tgba':
                            if vn == 'a' and i > 0:
                                continue
                            elif vn == 'a' and i == 0:
                                vnn = 'a'
                            else:
                                vnn = vn + str(i)
                            if vnn in self.pars.columns:
                                labstr_fit += pard[vn].format(self.pars.loc[ii, vnn])
                            elif fix is not None and vnn in fix.keys():
                                labstr_fit += 'fix '+ pard[vn].format(fix[vnn])
                            else:
                                labstr_fit += pard[vn].format(0)

            labstr_data = None
            if 'legd' in doplot:
                if data_label is not None:
                    labstr_data = data_label
                else:
                    labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'legq' in doplot:
                labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(self.qv[iq])

            if 'fit' in doplot and sumidit:
                pl.append(ax.plot(xf, msdf, '-', label=labstr_fit, linewidth=1))

            if 'data' in doplot:
                pl.append(ax.errorbar(self.t.filled(np.nan), self.msd.filled(np.nan), yerr=self.dmsd.filled(np.nan),
                                          linestyle='', marker=marker, label=labstr_data,
                                          alpha=alpha, markersize=markersize))

            if colors is None:
                if 'data' in doplot:
                    color = pl[-1].get_color()
                else:
                    color = 'gray'
            else:
                color = colors[ii]

            for p in pl:
                if isinstance(p, list):
                    p[0].set_color(color)
                elif isinstance(p, matplotlib.container.ErrorbarContainer):
                    for child in p.get_children():
                        child.set_color(color)

            ax.set_xscale('log')
            ax.set_yscale('log')

            if 'leg' in doplot:
                ax.legend()

            if 'report' in doplot and sumidit:
                print(lmfit.fit_report(out))

            ax.set_xlabel(r'$\tau\,(\mathrm{s})$')
            ax.set_ylabel(r'$g_2(\tau)$')

            ax.set_xlim(*xl)
            if yl is not None:
                ax.set_ylim(*yl)

        return None

    def _calc_model(self, v, x=None):
        """Calculate multi mode g2 funktion
        """
        t = self.t if x is None else x
        model = 6 * v['d']**2 \
                * (1 - np.exp(-(v['D0']/v['d']**2 * t)**v['a'])) ** (1/v['a']) \
                * (1 + v['Dm']/v['d']**2 * t) ** (v['a1'])
        if self._fit_logarithmic:
            return np.log10(model)
        else:
            return model

    def _residuals(self, pars, *args, **kwargs):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        model = self._calc_model(v)

        resid = np.sqrt((self._fit_data - model)**2 * np.abs(self._fit_weights)**2)
        return resid


    def _init_parameters(self, init, fix):
        '''Initialize lmfit parameters dictionary
        '''

        self._make_varnames()
        self._initial_guess(init)

        # initialize parameters
        pars = lmfit.Parameters()
        for vn, vinit in init.items():
            pars.add(vn, value=vinit[0], min=vinit[1], max=vinit[2], vary=1)


        self._init_pars_dataframe(init)

        # setting parameters fixed
        for vn in fix.keys():
            if vn in pars.keys():
                pars[vn].set(value=fix[vn], min=-np.inf, max=np.inf, vary=0)

        # print(pars)
        self._lmpars = pars

    def _get_weights(self, mode):
        '''
        mask data points to exclude them from the fit based on error bars and nan values
        and define weights for the fit
        '''
        dmsd = self.dmsd
        msd = self.msd.copy()
        if dmsd is not None and mode is not 'none':
            excerr = (dmsd.filled(0)<=0)
            wgt = dmsd.copy()
            wgt = np.ma.masked_where(excerr, wgt)
            if mode == 'semilogx':
                wgt = np.log10(wgt+1)
            elif mode == 'semilogx2':
                wgt = 1/np.log10(wgt/msd)
            elif mode == 'semilogx3':
                wgt = 1/(msd * np.log10(wgt/msd))
            elif (mode == 'equal') or (mode == 'none') or (mode == None):
                wgt = np.ones_like(wgt)
            elif mode == 'logt':
                wgt = 1/np.log10(1+self.t)
            elif mode == 't':
                wgt = 1/self.t
            elif mode == 'sig**2':
                wgt = 1/wgt**2
            elif mode == 'sig':
                wgt = 1/wgt
            elif mode == 'logsig':
                wgt = 1/np.log10(wgt)
            elif mode == 'data':
                wgt = 1/self.msd.copy()
            elif mode == 'loglog':
                self._fit_data = np.log10(self._fit_data)
                wgt = np.ma.masked_array(np.ones_like(self.msd))
                self._fit_logarithmic = True
            else:
                raise ValueError(f'Error mode {mode} not understood.')
        else:
            wgt = np.ma.masked_array(np.ones_like(self.msd))

        excdat = ~np.isfinite(self.msd) | wgt.mask
        self.msd = np.ma.masked_where(excdat, self.msd)

        self._weights = wgt

    def _init_pars_dataframe(self, init):
        """
        initialize pandas DataFrame for returning results
        """
        names = list(init.keys())
        cols = [names[i // 2] if (i + 1) % 2 else 'd' + names[i // 2]
                for i in range(len(names) * 2)]
        cols.extend(['chisqr', 'redchi', 'bic', 'aic'])
        self.pars = pd.DataFrame(columns=cols)

    def _initial_guess(self, init):
        """
        make initial guess for parameters if not provided by init
        """
        for vn in self._varnames:
            if vn not in init.keys():
                if vn == 'd':
                    init[vn] = (50, 0, None)
                elif vn == 'Dm':
                    init[vn] = (1e8, 0, None)
                elif vn == 'D0':
                    init[vn] = (1e8, 0, None)
                elif vn == 'a':
                    init[vn] = (1, 0, None)
                elif vn == 'a1':
                    init[vn] = (1, 0, None)


    def _make_varnames(self):
        """Make dict of varnames for easy handling in _calc_model
        """
        self._varnames = {'d': [],
                          'D0': [],
                          'Dm': [],
                          'a': [],
                          'a1':[],
                          }

    def _write_to_pars(self, out):
        """ Save fit results in self.pars variable
        """
        for k in self._varnames.keys():
            cond = k in ['__lnsigma']
            gof = np.hstack([out.chisqr, out.redchi, out.bic, out.aic])
            param_name = df_name = k
            value = out.params[param_name].value
            try:
                stderr = 1. * out.params[param_name].stderr
            except TypeError:
                stderr = np.nan
            self.pars.loc[0, df_name] = value
            self.pars.loc[0, 'd'+df_name] = stderr
            self.pars.iloc[0, -4:] = gof




